/* Copyright (C) 2009 Kentoku Shiba

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"
#include <mysql/plugin.h>
#include "vp_param.h"
#include "vp_err.h"
#include "vp_include.h"
#include "ha_vp.h"
#include "vp_table.h"

#define VP_CAN_BG_SEARCH (LL(1) << 37)
#define VP_CAN_BG_INSERT (LL(1) << 38)
#define VP_CAN_BG_UPDATE (LL(1) << 39)

extern handlerton *vp_hton_ptr;
#ifndef WITHOUT_VP_BG_ACCESS
extern pthread_attr_t vp_pt_attr;
#endif

ha_vp::ha_vp(
) : handler(vp_hton_ptr, NULL)
{
  DBUG_ENTER("ha_vp::ha_vp");
  DBUG_PRINT("info",("vp this=%x", this));
  share = NULL;
  part_tables = NULL;
  use_tables = NULL;
  work_bitmap = NULL;
  blob_buff = NULL;
  ref_length = 0;
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
  allocated_top_table_fields = 0;
#endif
  additional_table_flags = 0;
  DBUG_VOID_RETURN;
}

ha_vp::ha_vp(
  handlerton *hton,
  TABLE_SHARE *table_arg
) : handler(hton, table_arg)
{
  DBUG_ENTER("ha_vp::ha_vp");
  DBUG_PRINT("info",("vp this=%x", this));
  share = NULL;
  part_tables = NULL;
  use_tables = NULL;
  work_bitmap = NULL;
  blob_buff = NULL;
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
  allocated_top_table_fields = 0;
#endif
  additional_table_flags = 0;
  ref_length = sizeof(my_off_t);
  DBUG_VOID_RETURN;
}

static const char *ha_vp_exts[] = {
  NullS
};

const char **ha_vp::bas_ext() const
{
  return ha_vp_exts;
}

int ha_vp::open(
  const char* name,
  int mode,
  uint test_if_locked
) {
  int error_num, roop_count;
  THD *thd = ha_thd();
  DBUG_ENTER("ha_vp::open");
  DBUG_PRINT("info",("vp this=%x", this));

  if (!vp_get_share(name, table, thd, this, &error_num))
    goto error_get_share;
  thr_lock_data_init(&share->lock, &lock, NULL);

  if (!(part_tables = (TABLE_LIST*)
    my_multi_malloc(MYF(MY_WME),
      &part_tables, sizeof(TABLE_LIST) * share->table_count,
      &use_tables, sizeof(uchar) * share->use_tables_size,
      &use_tables2, sizeof(uchar) * share->use_tables_size,
      &use_tables3, sizeof(uchar) * share->use_tables_size,
      &select_ignore, sizeof(uchar) * share->use_tables_size,
      &select_ignore_with_lock, sizeof(uchar) * share->use_tables_size,
      &update_ignore, sizeof(uchar) * share->use_tables_size,
      &work_bitmap, sizeof(uchar) * ((table_share->fields + 7) / 8),
      &work_bitmap2, sizeof(uchar) * ((table_share->fields + 7) / 8),
      &work_bitmap3, sizeof(uchar) * ((table_share->fields + 7) / 8),
      &work_bitmap4, sizeof(uchar) * ((table_share->fields + 7) / 8),
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
      &top_table_field_for_childs, sizeof(Field **) * share->table_count,
#endif
#ifndef WITHOUT_VP_BG_ACCESS
      &bg_base, sizeof(VP_BG_BASE) * share->table_count,
#endif
      NullS))
  ) {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_tables_alloc;
  }
  memcpy(part_tables, share->part_tables,
    sizeof(TABLE_LIST) * share->table_count);
  memcpy(select_ignore, share->select_ignore,
    sizeof(uchar) * share->use_tables_size);
  memcpy(select_ignore_with_lock, share->select_ignore_with_lock,
    sizeof(uchar) * share->use_tables_size);
  memset((uchar *) update_ignore, 0,
    sizeof(uchar) * share->use_tables_size);

#ifdef WITH_PARTITION_STORAGE_ENGINE
  /* for table partitioning */
  if (table->child_l)
    *(table->child_last_l) = part_tables;
#endif

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    part_tables[roop_count].lock_type = (thr_lock_type) mode;
    part_tables[roop_count].parent_l = NULL;
    if (roop_count < share->table_count - 1)
      part_tables[roop_count].next_global = &part_tables[roop_count + 1];
#ifndef WITHOUT_VP_BG_ACCESS
    bg_base[roop_count].table_idx = roop_count;
    bg_base[roop_count].part_table = &part_tables[roop_count];
    bg_base[roop_count].parent = this;
    bg_base[roop_count].bg_init = FALSE;
    bg_base[roop_count].bg_caller_sync_wait = FALSE;
#endif
  }

#ifdef WITH_PARTITION_STORAGE_ENGINE
  /* for table partitioning */
  if (!table->child_l)
#endif
    table->child_l = part_tables;

  table->child_last_l = &part_tables[share->table_count - 1].next_global;

  DBUG_PRINT("info",("vp blob_fields=%d", table_share->blob_fields));
  if (table_share->blob_fields)
  {
    if (!(blob_buff = new String[table_share->fields]))
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_init_blob_buff;
    }
    for (roop_count = 0; roop_count < table_share->fields; roop_count++)
      blob_buff[roop_count].set_charset(table->field[roop_count]->charset());
  }

  if (reset())
  {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error;
  }

  DBUG_RETURN(0);

error:
  delete [] blob_buff;
error_init_blob_buff:
  my_free(part_tables, MYF(0));
  part_tables = NULL;
error_tables_alloc:
  vp_free_share(share);
  share = NULL;
error_get_share:
  DBUG_RETURN(error_num);
}

int ha_vp::close()
{
#ifndef WITHOUT_VP_BG_ACCESS
  int roop_count;
#endif
  THD *thd = ha_thd();
#ifndef WITHOUT_VP_BG_ACCESS
  VP_BG_BASE *base;
#endif
  DBUG_ENTER("ha_vp::close");
  DBUG_PRINT("info",("vp this=%x", this));

#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    base = &bg_base[roop_count];
    free_bg_thread(base);
  }
#endif
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
  if (allocated_top_table_fields)
  {
    my_free(top_table_field_for_childs[0], MYF(0));
    allocated_top_table_fields = 0;
  }
#endif
  if (blob_buff)
    delete [] blob_buff;
  if (part_tables)
  {
    my_free(part_tables, MYF(0));
    part_tables = NULL;
  }
  if (share)
  {
    vp_free_share(share);
    share = NULL;
  }
  table->child_l = NULL;

  DBUG_RETURN(0);
}

uint ha_vp::lock_count() const
{
  DBUG_ENTER("ha_vp::lock_count");
  DBUG_RETURN(table_lock_count);
}

THR_LOCK_DATA **ha_vp::store_lock(
  THD *thd,
  THR_LOCK_DATA **to,
  enum thr_lock_type lock_type
) {
  int roop_count;
  DBUG_ENTER("ha_vp::store_lock");
  DBUG_PRINT("info",("vp this=%x", this));
	sql_command = thd_sql_command(thd);
  DBUG_PRINT("info",("vp sql_command=%u", sql_command));
  lock_type_sto = lock_type;
  DBUG_PRINT("info",("vp lock_type=%u", lock_type));
  switch (sql_command)
  {
    case SQLCOM_SELECT:
      if (lock_type == TL_READ_WITH_SHARED_LOCKS)
        lock_mode = 1;
      else if (lock_type <= TL_READ_NO_INSERT)
        lock_mode = 0;
      else
        lock_mode = -1;
      break;
    case SQLCOM_CREATE_TABLE:
    case SQLCOM_UPDATE:
    case SQLCOM_INSERT:
    case SQLCOM_INSERT_SELECT:
    case SQLCOM_DELETE:
    case SQLCOM_LOAD:
    case SQLCOM_REPLACE:
    case SQLCOM_REPLACE_SELECT:
    case SQLCOM_DELETE_MULTI:
    case SQLCOM_UPDATE_MULTI:
      if (lock_type >= TL_READ && lock_type <= TL_READ_NO_INSERT)
        lock_mode = 1;
      else
        lock_mode = -1;
      break;
    default:
      lock_mode = -1;
      break;
  }

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    to = part_tables[roop_count].table->file->store_lock(thd, to, lock_type);
  }
  DBUG_RETURN(to);
}

int ha_vp::external_lock(
  THD *thd,
  int lock_type
) {
  int error_num, error_num2, roop_count;
  DBUG_ENTER("ha_vp::external_lock");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp lock_type=%x", lock_type));
  DBUG_PRINT("info",("vp thd->options=%x", (int) thd->options));
  lock_type_ext = lock_type;
  if (lock_type == F_WRLCK)
    update_request = TRUE;
  else
    update_request = FALSE;
  error_num = 0;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if ((error_num2 =
      part_tables[roop_count].table->file->ha_external_lock(thd, lock_type)))
    {
      if (lock_type != F_UNLCK)
        goto error_external_lock;
      else
        error_num = error_num2;
    }
  }
  DBUG_RETURN(error_num);

error_external_lock:
  while (roop_count)
  {
    roop_count--;
    part_tables[roop_count].table->file->ha_external_lock(thd, F_UNLCK);
  }
  DBUG_RETURN(error_num2);
}

int ha_vp::reset()
{
  int error_num, error_num2, roop_count;
  DBUG_ENTER("ha_vp::reset");
  DBUG_PRINT("info",("vp this=%x", this));
  bulk_insert = FALSE;
  init_bitmap = FALSE;
  child_table_idx = 0;
  error_num = 0;
  if (table->children_attached)
  {
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if ((error_num2 = part_tables[roop_count].table->file->ha_reset()))
        error_num = error_num2;
    }
  }
  DBUG_RETURN(error_num);
}

int ha_vp::extra(
  enum ha_extra_function operation
) {
  int error_num, error_num2, roop_count;
  TABLE_LIST *part_table;
  DBUG_ENTER("ha_vp::extra");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp operation=%d", (int) operation));
  switch (operation)
  {
    case HA_EXTRA_CACHE:
      if ((error_num =
        part_tables[child_table_idx].table->file->extra(operation)))
        DBUG_RETURN(error_num);
      break;
    case HA_EXTRA_KEYREAD:
    case HA_EXTRA_NO_KEYREAD:
      break;
    case HA_EXTRA_ATTACH_CHILDREN:
      table_lock_count = 0;
      additional_table_flags |=
        (VP_CAN_BG_SEARCH | VP_CAN_BG_INSERT | VP_CAN_BG_UPDATE);
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        part_table = &part_tables[roop_count];
        if ((error_num = part_table->table->file->extra(operation)))
          DBUG_RETURN(error_num);
        table_lock_count += part_table->table->file->lock_count();
        if (part_table->get_child_def_version() !=
          part_table->table->s->get_table_def_version())
        {
          share->reinit = TRUE;
          part_table->set_child_def_version(
            part_table->table->s->get_table_def_version());
        }
        additional_table_flags &= ~((ulonglong)
          (part_table->table->file->ha_table_flags() ^
            (VP_CAN_BG_SEARCH | VP_CAN_BG_INSERT | VP_CAN_BG_UPDATE)));
      }
      if (!share->init || share->reinit)
      {
        if ((error_num =
          vp_correspond_columns(table, share, table_share, part_tables)))
          DBUG_RETURN(error_num);
      }
      cached_table_flags = table_flags();
      break;
    case HA_EXTRA_DETACH_CHILDREN:
      if (table->children_attached)
      {
        error_num = 0;
        additional_table_flags &= ~((ulonglong)
          (VP_CAN_BG_SEARCH | VP_CAN_BG_INSERT | VP_CAN_BG_UPDATE));
        cached_table_flags = table_flags();
        for (roop_count = 0; roop_count < share->table_count; roop_count++)
        {
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
          part_tables[roop_count].table->file->clear_top_table_fields();
#endif
          if ((error_num2 =
            part_tables[roop_count].table->file->extra(operation)))
            error_num = error_num2;
        }
        table->children_attached = FALSE;
        if (error_num)
          DBUG_RETURN(error_num);
      }
      break;
    default:
      if (table->children_attached)
      {
        for (roop_count = 0; roop_count < share->table_count; roop_count++)
        {
          if ((error_num =
            part_tables[roop_count].table->file->extra(operation)))
            DBUG_RETURN(error_num);
        }
      }
      break;
  }
  DBUG_RETURN(0);
}

int ha_vp::extra_opt(
  enum ha_extra_function operation,
  ulong cachesize
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::extra_opt");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp operation=%d", (int) operation));
  DBUG_PRINT("info",("vp cachesize=%lu", cachesize));
  switch (operation)
  {
    case HA_EXTRA_CACHE:
      if ((error_num =
        part_tables[child_table_idx].table->file->extra_opt(operation,
          cachesize)))
        DBUG_RETURN(error_num);
      break;
    default:
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        if ((error_num =
          part_tables[roop_count].table->file->extra_opt(operation,
            cachesize)))
          DBUG_RETURN(error_num);
      }
      break;
  }
  DBUG_RETURN(0);
}

int ha_vp::index_init(
  uint idx,
  bool sorted
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::index_init");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp idx=%u", idx));
  active_index = idx;
  memset(use_tables, 0, sizeof(uchar) * share->use_tables_size);
  child_keyread = FALSE;
  single_table = FALSE;
  set_used_table = FALSE;
  child_table_idx = share->table_count;
  child_key_idx = MAX_KEY;
  if (
    share->zero_record_update_mode &&
    (lock_mode > 0 || lock_type_ext == F_WRLCK) &&
    (sql_command == SQLCOM_UPDATE || sql_command == SQLCOM_UPDATE_MULTI)
  ) {
    for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
    {
      table->read_set->bitmap[roop_count] |=
        share->cpy_clm_bitmap[roop_count];
      table->write_set->bitmap[roop_count] |=
        share->cpy_clm_bitmap[roop_count];
    }
  }
  memcpy(work_bitmap3, table->read_set->bitmap,
    sizeof(uchar) * share->bitmap_size);
  memcpy(work_bitmap4, table->write_set->bitmap,
    sizeof(uchar) * share->bitmap_size);

  if (
    (error_num = choose_child_index(idx, work_bitmap3, work_bitmap4,
      &child_table_idx, &child_key_idx)) ||
    (!single_table &&
    (error_num = choose_child_tables(work_bitmap3, work_bitmap4)))
  ) {
    DBUG_RETURN(error_num);
  }
  DBUG_PRINT("info",("vp child_table_idx=%d", child_table_idx));
  DBUG_PRINT("info",("vp child_key_idx=%d", child_key_idx));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      DBUG_PRINT("info",("vp table_count=%d", roop_count));
      if (roop_count == child_table_idx)
      {
        if (
          child_keyread &&
          (error_num =
            part_tables[roop_count].table->file->extra(HA_EXTRA_KEYREAD))
        )
          DBUG_RETURN(error_num);
        if ((error_num =
          part_tables[roop_count].table->file->ha_index_init(
            child_key_idx, sorted)))
          DBUG_RETURN(error_num);
      } else {
        if ((error_num =
          part_tables[roop_count].table->file->ha_index_init(
            part_tables[roop_count].table->s->primary_key, FALSE)))
          DBUG_RETURN(error_num);
      }
    }
  }
  DBUG_RETURN(0);
}

int ha_vp::index_end()
{
  int error_num, error_num2, roop_count;
  DBUG_ENTER("ha_vp::index_end");
  DBUG_PRINT("info",("vp this=%x", this));
  error_num = 0;
  active_index = MAX_KEY;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      if ((error_num2 =
        part_tables[roop_count].table->file->ha_index_end()))
        error_num = error_num2;
    }
  }
  DBUG_RETURN(error_num);
}

int ha_vp::index_read_map(
  uchar *buf,
  const uchar *key,
  key_part_map keypart_map,
  enum ha_rkey_function find_flag
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_read_map");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  child_key = create_child_key(key, child_key_different, keypart_map);
  if (
    (error_num = table2->file->index_read_map(
      table2->record[0], child_key, keypart_map, find_flag)) ||
    (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
  )
      DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(ptr_diff))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(index_next(buf));
  }
  DBUG_RETURN(0);
}

int ha_vp::index_read_last_map(
  uchar *buf,
  const uchar *key,
  key_part_map keypart_map
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_read_last_map");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  child_key = create_child_key(key, child_key_different, keypart_map);
  if (
    (error_num = table2->file->index_read_last_map(
      table2->record[0], child_key, keypart_map)) ||
    (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(ptr_diff))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(index_prev(buf));
  }
  DBUG_RETURN(0);
}

int ha_vp::index_next(
  uchar *buf
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_next");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  do {
    if (
      (error_num = table2->file->index_next(table2->record[0])) ||
      (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
    )
      DBUG_RETURN(error_num);
    if (
      !single_table &&
      (error_num = get_child_record_by_pk(ptr_diff)) &&
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
  } while (error_num);
  DBUG_RETURN(0);
}

int ha_vp::index_prev(
  uchar *buf
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_prev");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  do {
    if (
      (error_num = table2->file->index_prev(table2->record[0])) ||
      (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
    )
      DBUG_RETURN(error_num);
    if (
      !single_table &&
      (error_num = get_child_record_by_pk(ptr_diff)) &&
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
  } while (error_num);
  DBUG_RETURN(0);
}

int ha_vp::index_first(
  uchar *buf
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_first");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  if (
    (error_num = table2->file->index_first(table2->record[0])) ||
    (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(ptr_diff))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(index_next(buf));
  }
  DBUG_RETURN(0);
}

int ha_vp::index_last(
  uchar *buf
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_last");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  if (
    (error_num = table2->file->index_last(table2->record[0])) ||
    (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(ptr_diff))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(index_prev(buf));
  }
  DBUG_RETURN(0);
}

int ha_vp::index_next_same(
  uchar *buf,
  const uchar *key,
  uint keylen
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::index_next_same");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  if (
    (error_num = table2->file->index_next_same(table2->record[0], child_key,
      table2->key_info[child_key_idx].key_length)) ||
    (error_num = get_child_record_by_idx(child_table_idx, ptr_diff))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(ptr_diff))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(index_next(buf));
  }
  DBUG_RETURN(0);
}

int ha_vp::read_range_first(
  const key_range *start_key,
  const key_range *end_key,
  bool eq_range,
  bool sorted
) {
  int error_num;
  TABLE *table2;
  DBUG_ENTER("ha_vp::read_range_first");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp start_key=%x", start_key));
  DBUG_PRINT("info",("vp end_key=%x", end_key));
  if (start_key)
  {
    child_start_key.keypart_map = start_key->keypart_map;
    child_start_key.flag = start_key->flag;
    child_start_key.length = start_key->length;
    child_start_key.key = create_child_key(
      start_key->key, child_key_different, start_key->keypart_map);
  }
  if (end_key)
  {
    child_end_key.keypart_map = end_key->keypart_map;
    child_end_key.flag = end_key->flag;
    child_end_key.length = end_key->length;
    child_end_key.key =  create_child_key(
      end_key->key, child_end_key_different, end_key->keypart_map);
  }
  table2 = part_tables[child_table_idx].table;
  if (
    (error_num = table2->file->read_range_first(
      start_key ? &child_start_key : NULL,
      end_key ? &child_end_key : NULL, eq_range, sorted)) ||
    (error_num = get_child_record_by_idx(child_table_idx, 0))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(0))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(read_range_next());
  }
  DBUG_RETURN(0);
}

int ha_vp::read_range_next()
{
  int error_num;
  TABLE *table2;
  DBUG_ENTER("ha_vp::read_range_next");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  do {
    if (
      (error_num = table2->file->read_range_next()) ||
      (error_num = get_child_record_by_idx(child_table_idx, 0))
    )
      DBUG_RETURN(error_num);
    if (
      !single_table &&
      (error_num = get_child_record_by_pk(0)) &&
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
  } while (error_num);
  DBUG_RETURN(0);
}

int ha_vp::read_multi_range_first(
  KEY_MULTI_RANGE **found_range_p,
  KEY_MULTI_RANGE *ranges,
  uint range_count,
  bool sorted,
  HANDLER_BUFFER *buffer
) {
  int error_num;
  TABLE *table2;
  THD *thd = ha_thd();
  KEY_MULTI_RANGE *multi_range, *tmp_multi_range;
  uchar *keys;
  int multi_range_mode =
    THDVAR(thd, multi_range_mode) < 0 ?
    share->multi_range_mode :
    THDVAR(thd, multi_range_mode);
  DBUG_ENTER("ha_vp::read_multi_range_first");
  DBUG_PRINT("info",("vp this=%x", this));
  if (multi_range_mode == 0)
    DBUG_RETURN(handler::read_multi_range_first(
      found_range_p, ranges, range_count, sorted, buffer));

  multi_range_sorted = sorted;
  multi_range_buffer = buffer;
  if (vp_bit_is_set(share->need_converting, child_table_idx))
  {
    /* copy multi range for child */
    multi_range = (KEY_MULTI_RANGE *) sql_alloc(
      sizeof(KEY_MULTI_RANGE) * range_count);
    keys = (uchar *) sql_alloc(
      sizeof(uchar) * MAX_KEY_LENGTH * range_count * 2);
    memcpy(multi_range, ranges, sizeof(KEY_MULTI_RANGE) * range_count);
    for (
      tmp_multi_range = multi_range,
      multi_range_curr = ranges,
      multi_range_end = ranges + range_count;
      multi_range_curr < multi_range_end;
      tmp_multi_range++,
      multi_range_curr++
    ) {
      DBUG_PRINT("info",("vp multi_range->start_key.key=%x",
        multi_range->start_key.key));
      tmp_multi_range->start_key.key = create_child_key(
        multi_range_curr->start_key.key, keys,
        multi_range_curr->start_key.keypart_map);
      keys += MAX_KEY_LENGTH;
      DBUG_PRINT("info",("vp multi_range->end_key.key=%x",
        multi_range->start_key.key));
      tmp_multi_range->end_key.key = create_child_key(
        multi_range_curr->end_key.key, keys,
        multi_range_curr->end_key.keypart_map);
      keys += MAX_KEY_LENGTH;
    }
  } else
    multi_range = ranges;

  multi_range_curr = ranges;
  multi_range_end = ranges + range_count;

  child_found_range = multi_range;
  table2 = part_tables[child_table_idx].table;
  if (
    (error_num = table2->file->read_multi_range_first(
      &child_found_range, multi_range, range_count, sorted, buffer)) ||
    (error_num = get_child_record_by_idx(child_table_idx, 0))
  )
    DBUG_RETURN(error_num);
  if (
    !single_table &&
    (error_num = get_child_record_by_pk(0))
  ) {
    if (
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
    else
      DBUG_RETURN(read_multi_range_next(&child_found_range));
  }
  *found_range_p = multi_range - child_found_range + ranges;

  DBUG_RETURN(0);
}

int ha_vp::read_multi_range_next(
  KEY_MULTI_RANGE **found_range_p
) {
  int error_num;
  TABLE *table2;
  THD *thd = ha_thd();
  KEY_MULTI_RANGE *tmp_multi_range;
  int multi_range_mode =
    THDVAR(thd, multi_range_mode) < 0 ?
    share->multi_range_mode :
    THDVAR(thd, multi_range_mode);
  DBUG_ENTER("ha_vp::read_multi_range_next");
  DBUG_PRINT("info",("vp this=%x", this));
  if (multi_range_mode == 0)
    DBUG_RETURN(handler::read_multi_range_next(found_range_p));

  table2 = part_tables[child_table_idx].table;
  tmp_multi_range = child_found_range;
  do {
    if (
      (error_num = table2->file->read_multi_range_next(&child_found_range)) ||
      (error_num = get_child_record_by_idx(child_table_idx, 0))
    )
      DBUG_RETURN(error_num);
    if (
      !single_table &&
      (error_num = get_child_record_by_pk(0)) &&
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
  } while (error_num);
  *found_range_p = tmp_multi_range - child_found_range + *found_range_p;

  DBUG_RETURN(0);
}

int ha_vp::rnd_init(
  bool scan
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::rnd_init");
  DBUG_PRINT("info",("vp this=%x", this));
  if (scan)
  {
#ifndef DBUG_OFF
    for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
      DBUG_PRINT("info",("vp bitmap is %d-%d", roop_count,
        table->read_set->bitmap[roop_count]));
#endif
    memset(use_tables, 0, sizeof(uchar) * share->use_tables_size);
    child_keyread = FALSE;
    single_table = FALSE;
    set_used_table = FALSE;
    if (
      share->zero_record_update_mode &&
      (lock_mode > 0 || lock_type_ext == F_WRLCK) &&
      (sql_command == SQLCOM_UPDATE || sql_command == SQLCOM_UPDATE_MULTI)
    ) {
      for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
      {
        table->read_set->bitmap[roop_count] |=
          share->cpy_clm_bitmap[roop_count];
        table->write_set->bitmap[roop_count] |=
          share->cpy_clm_bitmap[roop_count];
      }
    }
    memcpy(work_bitmap3, table->read_set->bitmap,
      sizeof(uchar) * share->bitmap_size);
    memcpy(work_bitmap4, table->write_set->bitmap,
      sizeof(uchar) * share->bitmap_size);

    if (
      (error_num = choose_child_tables(work_bitmap3, work_bitmap4))
    ) {
      DBUG_RETURN(error_num);
    }
  }

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      DBUG_PRINT("info",("vp table_count=%d", roop_count));
      if (roop_count == child_table_idx)
      {
        if ((error_num =
          part_tables[roop_count].table->file->ha_rnd_init(scan)))
          DBUG_RETURN(error_num);
      } else {
        if ((error_num =
          part_tables[roop_count].table->file->ha_index_init(
            part_tables[roop_count].table->s->primary_key, FALSE)))
          DBUG_RETURN(error_num);
      }
    }
  }
  DBUG_RETURN(0);
}

int ha_vp::rnd_end()
{
  int error_num, error_num2, roop_count;
  DBUG_ENTER("ha_vp::rnd_end");
  DBUG_PRINT("info",("vp this=%x", this));
  error_num = 0;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      if (roop_count == child_table_idx)
      {
        if ((error_num2 =
          part_tables[roop_count].table->file->ha_rnd_end()))
          error_num = error_num2;
      } else {
        if ((error_num2 =
          part_tables[roop_count].table->file->ha_index_end()))
          error_num = error_num2;
      }
    }
  }
  DBUG_RETURN(error_num);
}

int ha_vp::rnd_next(
  uchar *buf
) {
  int error_num;
  TABLE *table2;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::rnd_next");
  DBUG_PRINT("info",("vp this=%x", this));
  table2 = part_tables[child_table_idx].table;
  do {
    if ((error_num = table2->file->rnd_next(table2->record[0])))
    {
      if (error_num == HA_ERR_RECORD_DELETED)
        continue;
      DBUG_RETURN(error_num);
    }
    if ((error_num = get_child_record_by_idx(child_table_idx, ptr_diff)))
      DBUG_RETURN(error_num);
    if (
      !single_table &&
      (error_num = get_child_record_by_pk(ptr_diff)) &&
      error_num != HA_ERR_KEY_NOT_FOUND &&
      error_num != HA_ERR_END_OF_FILE
    )
      DBUG_RETURN(error_num);
  } while (error_num);
  DBUG_RETURN(0);
}

void ha_vp::position(
  const uchar *record
) {
  int roop_count;
  TABLE *table2;
  handler *file2;
  uchar **pos_ptr;
  DBUG_ENTER("ha_vp::position");
  DBUG_PRINT("info",("vp this=%x", this));
  if (!(pos_ptr = (uchar **) sql_alloc(sizeof(uchar *) * share->table_count)))
  {
    ref = NULL;
    DBUG_VOID_RETURN;
  }
  DBUG_PRINT("info",("vp pos_ptr=%x", pos_ptr));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      table2 = part_tables[roop_count].table;
      file2 = table2->file;
      file2->position(table2->record[0]);
      pos_ptr[roop_count] = file2->ref;
    }
  }
  my_store_ptr(ref, ref_length, (my_off_t) pos_ptr);
  DBUG_VOID_RETURN;
}

int ha_vp::rnd_pos(
  uchar *buf,
  uchar *pos
) {
  int error_num, roop_count;
  TABLE *table2;
  uchar **pos_ptr;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  DBUG_ENTER("ha_vp::rnd_pos");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp pos=%x", pos));
  DBUG_PRINT("info",("vp buf=%x", buf));
  pos_ptr = (uchar **) my_get_ptr(pos, ref_length);
  DBUG_PRINT("info",("vp pos_ptr=%x", pos_ptr));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (vp_bit_is_set(use_tables, roop_count))
    {
      table2 = part_tables[roop_count].table;
      if (
        (error_num = table2->file->rnd_pos(table2->record[0],
          pos_ptr[roop_count])) ||
        (error_num = get_child_record_by_idx(roop_count, ptr_diff))
      )
        DBUG_RETURN(error_num);
    }
  }
  DBUG_RETURN(0);
}

int ha_vp::info(
  uint flag
) {
  int error_num, roop_count, roop_count2, info_src_table, table_idx, key_idx;
  handler *child_file;
  ha_statistics *child_stats;
  time_t create_time, check_time, update_time;
  ha_rows records;
  ulong mean_rec_length;
  uint block_size;
  ulonglong data_file_length, max_data_file_length, index_file_length,
    auto_increment_value;
  KEY *key_info, *key_info2;
  DBUG_ENTER("ha_vp::info");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp flag=%x", flag));

  if (table->children_attached)
  {
    memset(use_tables3, 0, sizeof(uchar) * share->use_tables_size);
    if (flag & HA_STATUS_ERRKEY)
    {
      child_file = part_tables[dup_table_idx].table->file;
      if ((error_num = child_file->info(flag)))
        DBUG_RETURN(error_num);
      vp_set_bit(use_tables3, dup_table_idx);

      errkey = child_file->errkey;
    }

    if (flag & (HA_STATUS_TIME | HA_STATUS_CONST | HA_STATUS_VARIABLE |
      HA_STATUS_AUTO))
    {
      if (share->info_src_table)
        info_src_table = share->info_src_table - 1;
      else
        info_src_table = 0;

      child_file = part_tables[info_src_table].table->file;
      if ((error_num = child_file->info(flag)))
        DBUG_RETURN(error_num);
      vp_set_bit(use_tables3, info_src_table);

      child_stats = &child_file->stats;
      update_time = child_stats->update_time;
      max_data_file_length = child_stats->max_data_file_length;
      create_time = child_stats->create_time;
      block_size = child_stats->block_size;
      data_file_length = child_stats->data_file_length;
      index_file_length = child_stats->index_file_length;
      records = child_stats->records;
      mean_rec_length = child_stats->mean_rec_length;
      check_time = child_stats->check_time;

      if (
        (flag & HA_STATUS_AUTO) &&
        info_src_table != share->auto_increment_table
      ) {
        child_file = part_tables[share->auto_increment_table].table->file;
        if ((error_num = child_file->info(flag)))
          DBUG_RETURN(error_num);
        vp_set_bit(use_tables3, share->auto_increment_table);
        child_stats = &child_file->stats;
        auto_increment_value = child_stats->auto_increment_value;
      } else
        auto_increment_value = child_stats->auto_increment_value;

      if (!share->info_src_table)
      {
        for (roop_count = 1; roop_count < share->table_count; roop_count++)
        {
          child_file = part_tables[roop_count].table->file;
          if ((error_num = child_file->info(flag)))
            DBUG_RETURN(error_num);
          vp_set_bit(use_tables3, roop_count);

          child_stats = &child_file->stats;
          if (difftime(child_stats->update_time, update_time) > 0)
            update_time = child_stats->update_time;
          max_data_file_length += child_stats->max_data_file_length;
          if (difftime(child_stats->create_time, create_time) > 0)
            create_time = child_stats->create_time;
          if (block_size < child_stats->block_size)
            block_size = child_stats->block_size;
          data_file_length += child_stats->data_file_length;
          index_file_length += child_stats->index_file_length;
          mean_rec_length += child_stats->mean_rec_length;
          if (difftime(child_stats->check_time, check_time) > 0)
            check_time = child_stats->check_time;
        }
      }
      if (thd_sql_command(ha_thd()) == SQLCOM_SHOW_KEYS)
      {
        for (roop_count = 0; roop_count < table_share->keys; roop_count++)
        {
          table_idx = share->keys[roop_count].correspond_key->table_idx;
          key_idx = share->keys[roop_count].correspond_key->key_idx;
          if (vp_bit_is_set(use_tables3, table_idx))
          {
            child_file = part_tables[table_idx].table->file;
            if ((error_num = child_file->info(flag)))
              DBUG_RETURN(error_num);
            vp_set_bit(use_tables3, table_idx);
          }
          key_info = &table->key_info[roop_count];
          key_info2 = &part_tables[table_idx].table->key_info[key_idx];
          for (roop_count2 = 0; roop_count2 < key_info->key_parts;
            roop_count2++)
            key_info->rec_per_key[roop_count2] =
              key_info2->rec_per_key[roop_count2];
        }
      }

      if (flag & HA_STATUS_TIME)
        stats.update_time = update_time;
      if (flag & HA_STATUS_CONST)
      {
        stats.max_data_file_length = max_data_file_length;
        stats.create_time = create_time;
        stats.block_size = block_size;
      }
      if (flag & HA_STATUS_VARIABLE)
      {
        stats.data_file_length = data_file_length;
        stats.index_file_length = index_file_length;
        stats.records = records;
        stats.mean_rec_length = mean_rec_length;
        stats.check_time = check_time;
      }
      if (flag & HA_STATUS_AUTO)
        stats.auto_increment_value = auto_increment_value;
    }
  }
  DBUG_RETURN(0);
}

ha_rows ha_vp::records_in_range(
  uint idx,
  key_range *start_key,
  key_range *end_key
) {
  int error_num, active_index_bak, roop_count;
  KEY *key_info, *key_info2;
  DBUG_ENTER("ha_vp::records_in_range");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp start_key=%x", start_key));
  DBUG_PRINT("info",("vp end_key=%x", end_key));
  memcpy(work_bitmap3, table->read_set->bitmap,
    sizeof(uchar) * share->bitmap_size);
  memcpy(work_bitmap4, table->write_set->bitmap,
    sizeof(uchar) * share->bitmap_size);
  if (
    (error_num = choose_child_index(idx, work_bitmap3, work_bitmap4,
      &child_table_idx, &child_key_idx))
  )
    DBUG_RETURN(error_num);

  active_index_bak = active_index;
  active_index = idx;
  if (start_key)
  {
    child_start_key.keypart_map = start_key->keypart_map;
    child_start_key.flag = start_key->flag;
    child_start_key.length = start_key->length;
    child_start_key.key = create_child_key(
      start_key->key, child_key_different, start_key->keypart_map);
  }
  if (end_key)
  {
    child_end_key.keypart_map = end_key->keypart_map;
    child_end_key.flag = end_key->flag;
    child_end_key.length = end_key->length;
    child_end_key.key =  create_child_key(
      end_key->key, child_end_key_different, end_key->keypart_map);
  }
  active_index = active_index_bak;
  if ((error_num = part_tables[child_table_idx].table->file->records_in_range(
    child_key_idx,
    start_key ? &child_start_key : NULL,
    end_key ? &child_end_key : NULL)))
    DBUG_RETURN(error_num);

  key_info = &table->key_info[idx];
  key_info2 = &part_tables[child_table_idx].table->key_info[child_key_idx];
  for (roop_count = 0; roop_count < key_info->key_parts; roop_count++)
    key_info->rec_per_key[roop_count] = key_info2->rec_per_key[roop_count];
  DBUG_RETURN(0);
}

const char *ha_vp::table_type() const
{
  DBUG_ENTER("ha_vp::table_type");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN("VP"); 
}

ulonglong ha_vp::table_flags() const
{
  DBUG_ENTER("ha_vp::table_flags");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(
    HA_REC_NOT_IN_SEQ |
    HA_CAN_GEOMETRY |
    HA_NULL_IN_KEY |
    HA_CAN_INDEX_BLOBS |
    HA_AUTO_PART_KEY |
    HA_REQUIRE_PRIMARY_KEY |
    /* HA_CAN_RTREEKEYS | */
    HA_PRIMARY_KEY_REQUIRED_FOR_DELETE |
    HA_NO_PREFIX_CHAR_KEYS |
    /* HA_CAN_FULLTEXT | */
    HA_CAN_SQL_HANDLER |
    HA_FILE_BASED |
    HA_CAN_INSERT_DELAYED |
    HA_CAN_BIT_FIELD |
    HA_NO_COPY_ON_ALTER |
    HA_BINLOG_ROW_CAPABLE |
    HA_BINLOG_STMT_CAPABLE |
    additional_table_flags |
    (share ? share->additional_table_flags : 0)
  );
}

const char *ha_vp::index_type(
  uint key_number
) {
  KEY *key_info;
  DBUG_ENTER("ha_vp::index_type");
  DBUG_PRINT("info",("vp this=%x", this));
  key_info = &table_share->key_info[key_number];
  DBUG_RETURN(
    (key_info->flags & HA_FULLTEXT) ? "FULLTEXT" :
    (key_info->flags & HA_SPATIAL) ? "SPATIAL" :
    (key_info->algorithm == HA_KEY_ALG_HASH) ? "HASH" :
    (key_info->algorithm == HA_KEY_ALG_RTREE) ? "RTREE" :
    "BTREE"
  );
}

ulong ha_vp::index_flags(
  uint idx,
  uint part,
  bool all_parts
) const {
  KEY *key_info;
  DBUG_ENTER("ha_vp::index_flags");
  DBUG_PRINT("info",("vp this=%x", this));
  key_info = &table_share->key_info[idx];
  DBUG_RETURN(
    (key_info->algorithm == HA_KEY_ALG_FULLTEXT) ?
      0 :
    (key_info->algorithm == HA_KEY_ALG_HASH) ?
      HA_ONLY_WHOLE_INDEX | HA_KEY_SCAN_NOT_ROR :
    HA_READ_NEXT | HA_READ_PREV | HA_READ_ORDER | HA_READ_RANGE |
    HA_KEYREAD_ONLY
  );
}

uint ha_vp::max_supported_record_length() const
{
  DBUG_ENTER("ha_vp::max_supported_record_length");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(HA_MAX_REC_LENGTH);
}

uint ha_vp::max_supported_keys() const
{
  DBUG_ENTER("ha_vp::max_supported_keys");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(MAX_KEY);
}

uint ha_vp::max_supported_key_parts() const
{
  DBUG_ENTER("ha_vp::max_supported_key_parts");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(MAX_REF_PARTS);
}

uint ha_vp::max_supported_key_length() const
{
  DBUG_ENTER("ha_vp::max_supported_key_length");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(VP_MAX_KEY_LENGTH);
}

uint ha_vp::max_supported_key_part_length() const
{
  DBUG_ENTER("ha_vp::max_supported_key_part_length");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(VP_MAX_KEY_LENGTH);
}

uint8 ha_vp::table_cache_type()
{
  DBUG_ENTER("ha_vp::table_cache_type");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(share->support_table_cache);
}

int ha_vp::update_auto_increment()
{
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::update_auto_increment");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(file->update_auto_increment());
}

void ha_vp::set_next_insert_id(
  ulonglong id
) {
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::set_next_insert_id");
  DBUG_PRINT("info",("vp this=%x", this));
  file->set_next_insert_id(id);
  DBUG_VOID_RETURN;
}

void ha_vp::get_auto_increment(
  ulonglong offset,
  ulonglong increment,
  ulonglong nb_desired_values,
  ulonglong *first_value,
  ulonglong *nb_reserved_values
) {
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::get_auto_increment");
  DBUG_PRINT("info",("vp this=%x", this));
  file->get_auto_increment(
    offset,
    increment,
    nb_desired_values,
    first_value,
    nb_reserved_values
  );
  DBUG_VOID_RETURN;
}

void ha_vp::restore_auto_increment(
  ulonglong prev_insert_id
) {
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::restore_auto_increment");
  DBUG_PRINT("info",("vp this=%x", this));
  file->restore_auto_increment(prev_insert_id);
  DBUG_VOID_RETURN;
}

void ha_vp::release_auto_increment()
{
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::release_auto_increment");
  DBUG_PRINT("info",("vp this=%x", this));
  file->ha_release_auto_increment();
  DBUG_VOID_RETURN;
}

int ha_vp::reset_auto_increment(
  ulonglong value
) {
  handler *file = part_tables[share->auto_increment_table].table->file;
  DBUG_ENTER("ha_vp::reset_auto_increment");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(file->ha_reset_auto_increment(value));
}

void ha_vp::start_bulk_insert(
  ha_rows rows
) {
  int roop_count;
  DBUG_ENTER("ha_vp::start_bulk_insert");
  DBUG_PRINT("info",("vp this=%x", this));
  bulk_insert = TRUE;
  if (!init_bitmap)
  {
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if (!(vp_bit_is_set(update_ignore, roop_count)))
      {
        clear_child_bitmap(roop_count);
        set_child_bitmap(
          (uchar *) table->write_set->bitmap,
          roop_count, TRUE);
      }
    }
    init_bitmap = TRUE;
  }
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
    part_tables[roop_count].table->file->ha_start_bulk_insert(rows);
  DBUG_VOID_RETURN;
}

int ha_vp::end_bulk_insert()
{
  int error_num, error_num2, roop_count;
  DBUG_ENTER("ha_vp::end_bulk_insert");
  DBUG_PRINT("info",("vp this=%x", this));
  bulk_insert = FALSE;
  error_num = 0;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      if ((error_num2 =
        part_tables[roop_count].table->file->ha_end_bulk_insert()))
        error_num = error_num2;
    }
  }
  DBUG_RETURN(error_num);
}

int ha_vp::write_row(
  uchar *buf
) {
  int error_num = 0, roop_count;
  THD *thd = table->in_use;
  TABLE *child_table;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  int child_binlog =
    THDVAR(thd, child_binlog) < 0 ?
    share->child_binlog :
    THDVAR(thd, child_binlog);
  ulonglong option_backup;
#ifndef WITHOUT_VP_BG_ACCESS
  int bgi_mode =
    THDVAR(thd, bgi_mode) < 0 ?
    share->bgi_mode :
    THDVAR(thd, bgi_mode);
  VP_BG_BASE *base;
#endif
  DBUG_ENTER("ha_vp::write_row");
  DBUG_PRINT("info",("vp this=%x", this));
  ha_statistic_increment(&SSV::ha_write_count);
  if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_INSERT)
    table->timestamp_field->set_time();

  if (!bulk_insert && !init_bitmap)
  {
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if (!(vp_bit_is_set(update_ignore, roop_count)))
      {
        clear_child_bitmap(roop_count);
        set_child_bitmap(
          (uchar *) table->write_set->bitmap,
          roop_count, TRUE);
      }
    }
    init_bitmap = TRUE;
  }

  memset(use_tables2, ~((uchar) 0), sizeof(uchar) * share->use_tables_size);
  set_child_record_for_update(ptr_diff, 0, TRUE, FALSE);
  if (!child_binlog)
  {
    option_backup = thd->options;
    thd->options &= ~OPTION_BIN_LOG;
  }
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      child_table = part_tables[roop_count].table;
#ifndef WITHOUT_VP_BG_ACCESS
      if (
        bgi_mode &&
        (child_table->file->ha_table_flags() & VP_CAN_BG_INSERT)
      ) {
        base = &bg_base[roop_count];
        if ((error_num = create_bg_thread(base)))
        {
          dup_table_idx = roop_count;
          if (!child_binlog)
            thd->options = option_backup;
          goto error;
        }
        base->bg_command = VP_BG_COMMAND_INSERT;
        bg_kick(base);
      } else {
#endif
        if ((error_num =
          child_table->file->ha_write_row(child_table->record[0])))
        {
          dup_table_idx = roop_count;
          if (!child_binlog)
            thd->options = option_backup;
          goto error;
        }
#ifndef WITHOUT_VP_BG_ACCESS
      }
#endif
    }
  }

#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    child_table = part_tables[roop_count].table;
    if (
      bgi_mode &&
      (child_table->file->ha_table_flags() & VP_CAN_BG_INSERT)
    ) {
      base = &bg_base[roop_count];
      bg_wait(base);
      if (base->bg_error)
      {
        dup_table_idx = roop_count;
        if (!child_binlog)
          thd->options = option_backup;
        error_num = base->bg_error;
        goto error;
      }
    }
  }
#endif

  if (!child_binlog)
    thd->options = option_backup;
  DBUG_RETURN(error_num);

error:
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    base = &bg_base[roop_count];
    if (base->bg_init)
      bg_wait(base);
  }
#endif
  DBUG_RETURN(error_num);
}

int ha_vp::update_row(
  const uchar *old_data,
  uchar *new_data
) {
  int error_num, roop_count;
  THD *thd = table->in_use;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(new_data, table->record[0]);
  my_ptrdiff_t ptr_diff2 = PTR_BYTE_DIFF(old_data, table->record[0]);
  int child_binlog =
    THDVAR(thd, child_binlog) < 0 ?
    share->child_binlog :
    THDVAR(thd, child_binlog);
  ulonglong option_backup;
  VP_KEY_COPY vp_key_copy;
  TABLE *table2;
  KEY *key_info;
  KEY_PART_INFO *key_part;
  Field *field;
  int bgu_mode;
  int bgi_mode;
  uchar insert_table[share->use_tables_size];
  DBUG_ENTER("ha_vp::update_row");
  DBUG_PRINT("info",("vp this=%x", this));
  memset(insert_table, 0, sizeof(uchar) * share->use_tables_size);
#ifndef WITHOUT_VP_BG_ACCESS
  bgu_mode =
    THDVAR(thd, bgu_mode) < 0 ?
    share->bgu_mode :
    THDVAR(thd, bgu_mode);
  bgi_mode =
    THDVAR(thd, bgi_mode) < 0 ?
    share->bgi_mode :
    THDVAR(thd, bgi_mode);
  VP_BG_BASE *base;
#endif
  ha_statistic_increment(&SSV::ha_update_count);
  if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
    table->timestamp_field->set_time();
  if (!init_bitmap)
  {
    memset(use_tables2, 0, sizeof(uchar) * share->use_tables_size);
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if (!(vp_bit_is_set(update_ignore, roop_count)))
      {
        clear_child_bitmap(roop_count);
        if (set_child_bitmap(
          (uchar *) table->write_set->bitmap,
          roop_count, TRUE))
          vp_set_bit(use_tables2, roop_count);
        set_child_bitmap(
          (uchar *) table->read_set->bitmap,
          roop_count, FALSE);
      }
    }
    init_bitmap = TRUE;
  }

  if (!child_binlog)
  {
    option_backup = thd->options;
    thd->options &= ~OPTION_BIN_LOG;
  }
  vp_key_copy.init = FALSE;
  vp_key_copy.mem_root_init = FALSE;

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    DBUG_PRINT("info",("vp table_count=%d", roop_count));
    if (
      vp_bit_is_set(use_tables2, roop_count) &&
      !(vp_bit_is_set(update_ignore, roop_count))
    ) {
      table2 = part_tables[roop_count].table;
      if (
        !vp_bit_is_set(use_tables, roop_count)
      ) {
        DBUG_PRINT("info",("vp call search_by_pk"));
        if ((error_num = search_by_pk_for_update(roop_count, 1, &vp_key_copy,
          ptr_diff2, bgu_mode)))
        {
          if (
            !share->zero_record_update_mode ||
            !vp_bit_is_set(select_ignore_with_lock, roop_count) ||
            (error_num != HA_ERR_KEY_NOT_FOUND &&
              error_num != HA_ERR_END_OF_FILE)
          ) {
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
          vp_set_bit(insert_table, roop_count);
          error_num = 0;
        }
      } else
        store_record(table2, record[1]);
    }
  }
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      vp_bit_is_set(use_tables2, roop_count) &&
      !vp_bit_is_set(use_tables, roop_count) &&
      !(vp_bit_is_set(update_ignore, roop_count))
    ) {
      table2 = part_tables[roop_count].table;
      if (
        (
          bgu_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
        ) ||
        (
          bgi_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_INSERT)
        )
      ) {
        base = &bg_base[roop_count];
        bg_wait(base);
        if (base->bg_error)
        {
          if (!child_binlog)
            thd->options = option_backup;
          error_num = base->bg_error;
          goto error;
        }
      }
    }
  }
#endif

  set_child_record_for_update(ptr_diff, 0, TRUE, FALSE);
  set_child_record_for_update(ptr_diff2, 1, FALSE, FALSE);
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    DBUG_PRINT("info",("vp table_count=%d", roop_count));
    if (
      vp_bit_is_set(use_tables2, roop_count) &&
      !(vp_bit_is_set(update_ignore, roop_count))
    ) {
      DBUG_PRINT("info",("vp call ha_update_row"));
      if (vp_bit_is_set(insert_table, roop_count))
      {
#ifndef WITHOUT_VP_BG_ACCESS
        table2 = part_tables[roop_count].table;
        if (
          bgi_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_INSERT)
        ) {
          base = &bg_base[roop_count];
          if ((error_num = create_bg_thread(base)))
          {
            dup_table_idx = roop_count;
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
          base->bg_command = VP_BG_COMMAND_INSERT;
          bg_kick(base);
        } else {
#endif
          if (
            (error_num =
              part_tables[roop_count].table->file->ha_write_row(
                part_tables[roop_count].table->record[0]))
          ) {
            dup_table_idx = roop_count;
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
#ifndef WITHOUT_VP_BG_ACCESS
        }
#endif
      } else {
#ifndef WITHOUT_VP_BG_ACCESS
        table2 = part_tables[roop_count].table;
        if (
          bgu_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
        ) {
          base = &bg_base[roop_count];
          if ((error_num = create_bg_thread(base)))
          {
            dup_table_idx = roop_count;
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
          base->bg_command = VP_BG_COMMAND_UPDATE;
          bg_kick(base);
        } else {
#endif
          if (
            (error_num =
              part_tables[roop_count].table->file->ha_update_row(
                part_tables[roop_count].table->record[1],
                part_tables[roop_count].table->record[0]))
          ) {
            dup_table_idx = roop_count;
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
#ifndef WITHOUT_VP_BG_ACCESS
        }
#endif
      }
    }
  }
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      vp_bit_is_set(use_tables2, roop_count) &&
      !(vp_bit_is_set(update_ignore, roop_count))
    ) {
      table2 = part_tables[roop_count].table;
      if (
        (
          bgu_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
        ) ||
        (
          bgi_mode &&
          (table2->file->ha_table_flags() & VP_CAN_BG_INSERT)
        )
      ) {
        base = &bg_base[roop_count];
        bg_wait(base);
        if (base->bg_error)
        {
          if (!child_binlog)
            thd->options = option_backup;
          error_num = base->bg_error;
          goto error;
        }
      }
    }
  }
#endif

  if (!child_binlog)
    thd->options = option_backup;
  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(0);

error:
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      base = &bg_base[roop_count];
      if (base->bg_init)
        bg_wait(base);
    }
  }
#endif
  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(error_num);
}

int ha_vp::delete_row(
  const uchar *buf
) {
  int error_num, roop_count;
  THD *thd = table->in_use;
  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
  int child_binlog =
    THDVAR(thd, child_binlog) < 0 ?
    share->child_binlog :
    THDVAR(thd, child_binlog);
  ulonglong option_backup;
  VP_KEY_COPY vp_key_copy;
  KEY *key_info;
  TABLE *table2;
  int bgu_mode;
  DBUG_ENTER("ha_vp::delete_row");
  DBUG_PRINT("info",("vp this=%x", this));
#ifndef WITHOUT_VP_BG_ACCESS
  bgu_mode =
    THDVAR(thd, bgu_mode) < 0 ?
    share->bgu_mode :
    THDVAR(thd, bgu_mode);
  VP_BG_BASE *base;
#endif
  ha_statistic_increment(&SSV::ha_delete_count);
  if (!init_bitmap)
  {
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if (!(vp_bit_is_set(update_ignore, roop_count)))
      {
        clear_child_bitmap(roop_count);
        set_child_bitmap(
          (uchar *) table->read_set->bitmap,
          roop_count, FALSE);
      }
    }
    init_bitmap = TRUE;
  }

  memset(use_tables2, ~((uchar) 0), sizeof(uchar) * share->use_tables_size);
  set_child_record_for_update(ptr_diff, 0, FALSE, FALSE);
  if (!child_binlog)
  {
    option_backup = thd->options;
    thd->options &= ~OPTION_BIN_LOG;
  }
  vp_key_copy.init = FALSE;
  vp_key_copy.mem_root_init = FALSE;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      if (!vp_bit_is_set(use_tables, roop_count))
      {
        if ((error_num = search_by_pk_for_update(roop_count, 0, &vp_key_copy,
          ptr_diff, bgu_mode)))
        {
          if (!child_binlog)
            thd->options = option_backup;
          goto error;
        }
      }
      table2 = part_tables[roop_count].table;
#ifndef WITHOUT_VP_BG_ACCESS
      if (
        bgu_mode &&
        (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
      ) {
        if (vp_bit_is_set(use_tables, roop_count))
        {
          base = &bg_base[roop_count];
          if ((error_num = create_bg_thread(base)))
          {
            if (!child_binlog)
              thd->options = option_backup;
            goto error;
          }
          base->bg_command = VP_BG_COMMAND_DELETE;
          bg_kick(base);
        }
      } else {
#endif
        if ((error_num = table2->file->ha_delete_row(table2->record[0])))
        {
          if (!child_binlog)
            thd->options = option_backup;
          goto error;
        }
#ifndef WITHOUT_VP_BG_ACCESS
      }
#endif
    }
  }
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      table2 = part_tables[roop_count].table;
      if (
        bgu_mode &&
        (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
      ) {
        base = &bg_base[roop_count];
        bg_wait(base);
        if (base->bg_error)
        {
          if (!child_binlog)
            thd->options = option_backup;
          error_num = base->bg_error;
          goto error;
        }
      }
    }
  }
#endif

  if (!child_binlog)
    thd->options = option_backup;
  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(0);

error:
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      base = &bg_base[roop_count];
      if (base->bg_init)
        bg_wait(base);
    }
  }
#endif
  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(error_num);
}

int ha_vp::delete_all_rows()
{
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::delete_all_rows");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (!(vp_bit_is_set(update_ignore, roop_count)))
    {
      if ((error_num =
        part_tables[roop_count].table->file->ha_delete_all_rows()))
        DBUG_RETURN(error_num);
    }
  }
  DBUG_RETURN(0);
}

double ha_vp::scan_time()
{
  DBUG_ENTER("ha_vp::scan_time");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(part_tables[child_table_idx].table->file->scan_time());
}

double ha_vp::read_time(
  uint index,
  uint ranges,
  ha_rows rows
) {
  int tgt_table_idx;
  DBUG_ENTER("ha_vp::read_time");
  DBUG_PRINT("info",("vp this=%x", this));
  tgt_table_idx = share->keys->shortest_correspond_key->table_idx;
  DBUG_RETURN(part_tables[tgt_table_idx].table->file->scan_time());
}

const key_map *ha_vp::keys_to_use_for_scanning()
{
  DBUG_ENTER("ha_vp::keys_to_use_for_scanning");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(&key_map_full);
}

ha_rows ha_vp::estimate_rows_upper_bound()
{
  DBUG_ENTER("ha_vp::estimate_rows_upper_bound");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(
    part_tables[child_table_idx].table->file->estimate_rows_upper_bound());
}

bool ha_vp::get_error_message(
  int error,
  String *buf
) {
  DBUG_ENTER("ha_vp::get_error_message");
  DBUG_PRINT("info",("vp this=%x", this));
  if (buf->reserve(ER_VP_UNKNOWN_LEN))
    DBUG_RETURN(TRUE);
  buf->q_append(ER_VP_UNKNOWN_STR, ER_VP_UNKNOWN_LEN);
  DBUG_RETURN(FALSE);
}

int ha_vp::create(
  const char *name,
  TABLE *form,
  HA_CREATE_INFO *info
) {
  int error_num;
  VP_SHARE tmp_share;
  DBUG_ENTER("ha_vp::create");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp name=%s", name));
  DBUG_PRINT("info",
    ("vp form->s->connect_string=%s", form->s->connect_string.str));
  DBUG_PRINT("info",
    ("vp info->connect_string=%s", info->connect_string.str));
  memset(&tmp_share, 0, sizeof(VP_SHARE));
  tmp_share.table_name = (char*) name;
  tmp_share.table_name_length = strlen(name);
  if ((error_num = vp_parse_table_info(&tmp_share, form, 1)))
    goto error;

  vp_free_share_alloc(&tmp_share);
  DBUG_RETURN(0);

error:
  vp_free_share_alloc(&tmp_share);
  DBUG_RETURN(error_num);
}

void ha_vp::update_create_info(
  HA_CREATE_INFO* create_info
) {
  DBUG_ENTER("ha_vp::update_create_info");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",
    ("vp create_info->connect_string=%s",
    create_info->connect_string.str));
  DBUG_VOID_RETURN;
}

int ha_vp::rename_table(
  const char *from,
  const char *to
) {
  DBUG_ENTER("ha_vp::rename_table");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp from=%s", from));
  DBUG_PRINT("info",("vp to=%s", to));
  DBUG_RETURN(0);
}

int ha_vp::delete_table(
  const char *name
) {
  DBUG_ENTER("ha_vp::delete_table");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_PRINT("info",("vp name=%s", name));
  DBUG_RETURN(0);
}

bool ha_vp::is_crashed() const
{
  int roop_count;
  DBUG_ENTER("ha_vp::is_crashed");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (part_tables[roop_count].table->file->is_crashed())
      DBUG_RETURN(TRUE);
  }
  DBUG_RETURN(FALSE);
}

bool ha_vp::auto_repair() const
{
  DBUG_ENTER("ha_vp::auto_repair");
  DBUG_PRINT("info",("vp this=%x", this));
  DBUG_RETURN(FALSE);
}

int ha_vp::disable_indexes(
  uint mode
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::disable_indexes");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if ((error_num =
      part_tables[roop_count].table->file->ha_disable_indexes(mode)))
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(0);
}

int ha_vp::enable_indexes(
  uint mode
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::enable_indexes");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if ((error_num =
      part_tables[roop_count].table->file->ha_enable_indexes(mode)))
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(0);
}


int ha_vp::check(
  THD* thd,
  HA_CHECK_OPT* check_opt
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::check");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      (error_num =
        part_tables[roop_count].table->file->ha_check(thd, check_opt)) &&
      error_num != HA_ADMIN_ALREADY_DONE
    )
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(error_num);
}

int ha_vp::repair(
  THD* thd,
  HA_CHECK_OPT* check_opt
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::repair");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      (error_num =
        part_tables[roop_count].table->file->ha_repair(thd, check_opt)) &&
      error_num != HA_ADMIN_ALREADY_DONE
    )
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(error_num);
}

bool ha_vp::check_and_repair(
  THD *thd
) {
  int roop_count;
  DBUG_ENTER("ha_vp::check_and_repair");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (part_tables[roop_count].table->file->ha_check_and_repair(thd))
      DBUG_RETURN(TRUE);
  }
  DBUG_RETURN(FALSE);
}

int ha_vp::analyze(
  THD* thd,
  HA_CHECK_OPT* check_opt
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::analyze");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      (error_num =
        part_tables[roop_count].table->file->ha_analyze(thd, check_opt)) &&
      error_num != HA_ADMIN_ALREADY_DONE
    )
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(error_num);
}

int ha_vp::optimize(
  THD* thd,
  HA_CHECK_OPT* check_opt
) {
  int error_num, roop_count;
  DBUG_ENTER("ha_vp::optimize");
  DBUG_PRINT("info",("vp this=%x", this));
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      (error_num =
        part_tables[roop_count].table->file->ha_optimize(thd, check_opt)) &&
      error_num != HA_ADMIN_ALREADY_DONE
    )
      DBUG_RETURN(error_num);
  }
  DBUG_RETURN(error_num);
}

#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
int ha_vp::set_top_table_and_fields(
  TABLE *top_table,
  Field **top_table_field,
  uint top_table_fields,
  bool self
) {
  int roop_count, roop_count2;
  uint field_index, field_index2;
  DBUG_ENTER("ha_vp::set_top_table_fields");
  DBUG_PRINT("info",("vp this=%x", this));
  if (!set_top_table_fields || self != top_table_self)
  {
    if (top_table_fields > allocated_top_table_fields)
    {
      if (allocated_top_table_fields)
        my_free(top_table_field_for_childs[0], MYF(0));

      if (!(top_table_field_for_childs[0] =
        (Field **) my_malloc(sizeof(Field *) * (top_table_fields + 1) *
          share->table_count, MYF(MY_WME))))
        DBUG_RETURN(HA_ERR_OUT_OF_MEM);

      for (roop_count = 0; roop_count < share->table_count - 1; roop_count++)
        top_table_field_for_childs[roop_count + 1] =
          top_table_field_for_childs[roop_count] + top_table_fields;

      allocated_top_table_fields = top_table_fields;
    }

    for (roop_count = 0; roop_count < top_table_fields + 1; roop_count++)
    {
      if (top_table_field[roop_count])
      {
        field_index = top_table_field[roop_count]->field_index;
        for (roop_count2 = 0; roop_count2 < share->table_count; roop_count2++)
        {
          field_index2 = share->correspond_columns_p[table_share->fields *
            roop_count2 + field_index];
          if (field_index2 < MAX_FIELDS)
            (top_table_field_for_childs[roop_count2])[roop_count] =
              part_tables[roop_count2].table->field[field_index2];
          else
            (top_table_field_for_childs[roop_count2])[roop_count] = NULL;
        }
      } else {
        for (roop_count2 = 0; roop_count2 < share->table_count; roop_count2++)
          (top_table_field_for_childs[roop_count2])[roop_count] = NULL;
      }
    }

    set_top_table_fields = TRUE;
    this->top_table = top_table;
    this->top_table_field = top_table_field;
    this->top_table_fields = top_table_fields;
    top_table_self = self;
  }
  DBUG_RETURN(0);
}

int ha_vp::set_top_table_and_fields(
  TABLE *top_table,
  Field **top_table_field,
  uint top_table_fields
) {
  DBUG_ENTER("ha_vp::set_top_table_fields");
  DBUG_RETURN(set_top_table_and_fields(top_table, top_table_field,
    top_table_fields, FALSE));
}

const COND *ha_vp::cond_push(const COND *cond)
{
  int roop_count;
  COND *res_cond = NULL;
  DBUG_ENTER("ha_vp::cond_push");

  if (
    !set_top_table_fields &&
    set_top_table_and_fields(table, table->field, table_share->fields, TRUE)
  )
    DBUG_RETURN(cond);

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (part_tables[roop_count].table->file->set_top_table_and_fields(
      top_table,
      top_table_field_for_childs[roop_count],
      top_table_fields))
      DBUG_RETURN(cond);
  }

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (part_tables[roop_count].table->file->pushed_cond != cond)
    {
      if (part_tables[roop_count].table->file->cond_push(cond))
        res_cond = (COND *) cond;
      else
        part_tables[roop_count].table->file->pushed_cond = cond;
    }
  }
  DBUG_RETURN(res_cond);
}

void ha_vp::cond_pop()
{
  int roop_count;
  DBUG_ENTER("ha_vp::cond_push");

  for (roop_count = 0; roop_count < share->table_count; roop_count++)
    part_tables[roop_count].table->file->cond_pop();
  DBUG_VOID_RETURN;
}
#endif

int ha_vp::choose_child_index(
  uint idx,
  uchar *read_set,
  uchar *write_set,
  int *table_idx,
  int *key_idx
) {
  int error_num, roop_count, correspond_count, correspond_count2,
    choose_table_mode;
  uint field_count;
  uchar *tmp_columns_bit, *pk_columns_bit, *tmp_select_ignore;
  VP_CORRESPOND_KEY *tmp_correspond_key, *tmp_correspond_key2;
  bool correspond_flag, correspond_flag2;
  DBUG_ENTER("ha_vp::choose_child_index");

  if (lock_mode > 0 || lock_type_ext == F_WRLCK)
  {
    choose_table_mode =
      (THDVAR(ha_thd(), choose_table_mode_for_lock) < 0 ?
      share->choose_table_mode_for_lock :
      THDVAR(ha_thd(), choose_table_mode_for_lock));
    DBUG_PRINT("info",("vp choose_table_mode_for_lock=%d",
      choose_table_mode));
    tmp_select_ignore = select_ignore_with_lock;
  } else {
    choose_table_mode =
      (THDVAR(ha_thd(), choose_table_mode) < 0 ?
      share->choose_table_mode : THDVAR(ha_thd(), choose_table_mode));
    DBUG_PRINT("info",("vp choose_table_mode=%d",
      choose_table_mode));
    tmp_select_ignore = select_ignore;
  }

  /* choose a index phase 1 */
  correspond_flag = TRUE;
  tmp_columns_bit = share->keys[idx].columns_bit;
  for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
  {
    if (
      (read_set[roop_count] & tmp_columns_bit[roop_count]) !=
        read_set[roop_count] ||
      (write_set[roop_count] & tmp_columns_bit[roop_count]) !=
        write_set[roop_count]
    ) {
      correspond_flag = FALSE;
      break;
    }
  }
  if (correspond_flag)
  {
    if (choose_table_mode == 0)
    {
      tmp_correspond_key = share->keys[idx].shortest_correspond_key;
      while (vp_bit_is_set(tmp_select_ignore, tmp_correspond_key->table_idx))
      {
        if (!(tmp_correspond_key = tmp_correspond_key->next_shortest))
        {
          my_printf_error(ER_VP_IGNORED_CORRESPOND_KEY_NUM,
            ER_VP_IGNORED_CORRESPOND_KEY_STR, MYF(0), idx);
          DBUG_RETURN(ER_VP_IGNORED_CORRESPOND_KEY_NUM);
        }
      }
    } else {
      tmp_correspond_key = share->keys[idx].correspond_key;
      while (vp_bit_is_set(tmp_select_ignore, tmp_correspond_key->table_idx))
      {
        if (!(tmp_correspond_key = tmp_correspond_key->next))
        {
          my_printf_error(ER_VP_IGNORED_CORRESPOND_KEY_NUM,
            ER_VP_IGNORED_CORRESPOND_KEY_STR, MYF(0), idx);
          DBUG_RETURN(ER_VP_IGNORED_CORRESPOND_KEY_NUM);
        }
      }
    }
    *table_idx = tmp_correspond_key->table_idx;
    *key_idx = tmp_correspond_key->key_idx;
    child_keyread = TRUE;
    single_table = TRUE;
    set_used_table = TRUE;
    clear_child_bitmap(*table_idx);
    set_child_bitmap(read_set, *table_idx, FALSE);
    set_child_bitmap(write_set, *table_idx, TRUE);
    vp_set_bit(use_tables, *table_idx);
    memset(read_set, 0, sizeof(uchar) * share->bitmap_size);
    memset(write_set, 0, sizeof(uchar) * share->bitmap_size);
    DBUG_RETURN(0);
  }

  /* choose a index phase 2 */
  if (choose_table_mode == 0)
  {
    correspond_count = 0;
    field_count = 0;
    tmp_correspond_key = NULL;
    tmp_correspond_key2 = share->keys[idx].correspond_key;
    while (vp_bit_is_set(tmp_select_ignore, tmp_correspond_key2->table_idx))
    {
      if (!(tmp_correspond_key2 = tmp_correspond_key2->next))
      {
        my_printf_error(ER_VP_IGNORED_CORRESPOND_KEY_NUM,
          ER_VP_IGNORED_CORRESPOND_KEY_STR, MYF(0), idx);
        DBUG_RETURN(ER_VP_IGNORED_CORRESPOND_KEY_NUM);
      }
    }
    while (tmp_correspond_key2)
    {
      correspond_count2 = 0;
      tmp_columns_bit = tmp_correspond_key2->columns_bit;
      for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
        correspond_count2 += vp_bit_count(
          (read_set[roop_count] | write_set[roop_count]) &
          tmp_columns_bit[roop_count]);
      if (correspond_count < correspond_count2)
      {
        tmp_correspond_key = tmp_correspond_key2;
        correspond_count = correspond_count2;
        field_count =
          part_tables[tmp_correspond_key->table_idx].table->s->fields;
      } else if (
        correspond_count == correspond_count2 &&
        field_count >
          part_tables[tmp_correspond_key2->table_idx].table->s->fields
      ) {
        tmp_correspond_key = tmp_correspond_key2;
        field_count =
          part_tables[tmp_correspond_key->table_idx].table->s->fields;
      }
      tmp_correspond_key2 = tmp_correspond_key2->next;
      if (tmp_correspond_key2)
      {
        while (vp_bit_is_set(tmp_select_ignore,
          tmp_correspond_key2->table_idx))
        {
          if (!(tmp_correspond_key2 = tmp_correspond_key2->next))
            break;
        }
      }
    }
  } else {
    tmp_correspond_key = share->keys[idx].correspond_key;
    while (vp_bit_is_set(tmp_select_ignore, tmp_correspond_key->table_idx))
    {
      if (!(tmp_correspond_key = tmp_correspond_key->next))
      {
        my_printf_error(ER_VP_IGNORED_CORRESPOND_KEY_NUM,
          ER_VP_IGNORED_CORRESPOND_KEY_STR, MYF(0), idx);
        DBUG_RETURN(ER_VP_IGNORED_CORRESPOND_KEY_NUM);
      }
    }
  }

  /* choose a index phase 3 */
  correspond_flag = TRUE;
  tmp_columns_bit = tmp_correspond_key->columns_bit;
  for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
  {
    work_bitmap[roop_count] =
      read_set[roop_count] & tmp_columns_bit[roop_count];
    DBUG_PRINT("info",("vp work_bitmap[%d]=%d",
      roop_count, work_bitmap[roop_count]));
    read_set[roop_count] &= ~work_bitmap[roop_count];
    DBUG_PRINT("info",("vp read_set[%d]=%d",
      roop_count, read_set[roop_count]));

    work_bitmap2[roop_count] =
      write_set[roop_count] & tmp_columns_bit[roop_count];
    DBUG_PRINT("info",("vp work_bitmap2[%d]=%d",
      roop_count, work_bitmap2[roop_count]));
    write_set[roop_count] &= ~work_bitmap2[roop_count];
    DBUG_PRINT("info",("vp write_set[%d]=%d",
      roop_count, write_set[roop_count]));

    if (read_set[roop_count] || write_set[roop_count])
      correspond_flag = FALSE;
  }
  *table_idx = tmp_correspond_key->table_idx;
  *key_idx = tmp_correspond_key->key_idx;
  if (correspond_flag)
  {
    child_keyread = TRUE;
    single_table = TRUE;
  } else {
    /* with table scanning */
    correspond_flag = TRUE;
    correspond_flag2 = TRUE;
    tmp_columns_bit =
      &share->correspond_columns_bit[*table_idx * share->bitmap_size];
    pk_columns_bit =
      share->keys[table_share->primary_key].columns_bit;
    for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
    {
      if (
        correspond_flag2 &&
        (
          ((read_set[roop_count] | write_set[roop_count]) &
            tmp_columns_bit[roop_count]) ||
          ((work_bitmap[roop_count] | work_bitmap2[roop_count]) &
            pk_columns_bit[roop_count]) != pk_columns_bit[roop_count]
        )
      ) {
        correspond_flag2 = FALSE;
      }
      work_bitmap[roop_count] |=
        read_set[roop_count] & tmp_columns_bit[roop_count];
      DBUG_PRINT("info",("vp work_bitmap[%d]=%d",
        roop_count, work_bitmap[roop_count]));
      read_set[roop_count] &= ~work_bitmap[roop_count];
      DBUG_PRINT("info",("vp read_set[%d]=%d",
        roop_count, read_set[roop_count]));

      work_bitmap2[roop_count] |=
        write_set[roop_count] & tmp_columns_bit[roop_count];
      DBUG_PRINT("info",("vp work_bitmap2[%d]=%d",
        roop_count, work_bitmap2[roop_count]));
      write_set[roop_count] &= ~work_bitmap2[roop_count];
      DBUG_PRINT("info",("vp write_set[%d]=%d",
        roop_count, write_set[roop_count]));

      if (read_set[roop_count] || write_set[roop_count])
        correspond_flag = FALSE;
    }
    if (correspond_flag)
      single_table = TRUE;
    else if (correspond_flag2)
      child_keyread = TRUE;
  }
  set_used_table = TRUE;
  clear_child_bitmap(*table_idx);
  set_child_bitmap(work_bitmap, *table_idx, FALSE);
  set_child_bitmap(work_bitmap2, *table_idx, TRUE);
  vp_set_bit(use_tables, *table_idx);

  DBUG_RETURN(0);
}

int ha_vp::choose_child_tables(
  uchar *read_set,
  uchar *write_set
) {
  int error_num, roop_count, roop_count2, correspond_count, correspond_count2,
    table_idx, table_idx2, choose_table_mode;
  uint field_count;
  uchar *tmp_columns_bit, *tmp_select_ignore;
  bool correspond_flag;
  DBUG_ENTER("ha_vp::choose_child_tables");

  if (lock_mode > 0 || lock_type_ext == F_WRLCK)
  {
    choose_table_mode =
      (THDVAR(ha_thd(), choose_table_mode_for_lock) < 0 ?
      share->choose_table_mode_for_lock :
      THDVAR(ha_thd(), choose_table_mode_for_lock));
    DBUG_PRINT("info",("vp choose_table_mode_for_lock=%d",
      choose_table_mode));
    tmp_select_ignore = select_ignore_with_lock;
  } else {
    choose_table_mode =
      (THDVAR(ha_thd(), choose_table_mode) < 0 ?
      share->choose_table_mode : THDVAR(ha_thd(), choose_table_mode));
    DBUG_PRINT("info",("vp choose_table_mode=%d",
      choose_table_mode));
    tmp_select_ignore = select_ignore;
  }
  table_idx2 = 0;
  do {
    table_idx = share->table_count;
    if (choose_table_mode == 0)
    {
      correspond_count = 0;
      field_count = 0;
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        if (
          !(vp_bit_is_set(use_tables, roop_count)) &&
          !(vp_bit_is_set(tmp_select_ignore, roop_count))
        ) {
          correspond_count2 = 0;
          tmp_columns_bit =
            &share->correspond_columns_bit[roop_count * share->bitmap_size];
          for (roop_count2 = 0; roop_count2 < share->bitmap_size;
            roop_count2++)
            correspond_count2 += vp_bit_count(
              (read_set[roop_count2] | write_set[roop_count2]) &
              tmp_columns_bit[roop_count2]);
          if (correspond_count < correspond_count2)
          {
            table_idx = roop_count;
            correspond_count = correspond_count2;
            field_count =
              part_tables[roop_count].table->s->fields;
          } else if (
            correspond_count == correspond_count2 &&
            field_count >
              part_tables[roop_count].table->s->fields
          ) {
            table_idx = roop_count;
            field_count =
              part_tables[roop_count].table->s->fields;
          }
        }
      }
    } else {
      for (roop_count = table_idx2; roop_count < share->table_count;
        roop_count++)
      {
        if (
          !(vp_bit_is_set(use_tables, roop_count)) &&
          !(vp_bit_is_set(tmp_select_ignore, roop_count))
        ) {
          tmp_columns_bit =
            &share->correspond_columns_bit[roop_count * share->bitmap_size];
          for (roop_count2 = 0; roop_count2 < share->bitmap_size;
            roop_count2++)
          {
            if (vp_bit_count(
              (read_set[roop_count2] | write_set[roop_count2]) &
              tmp_columns_bit[roop_count2]))
              break;
          }
          if (roop_count2 < share->bitmap_size)
          {
            table_idx2 = table_idx = roop_count;
            break;
          }
        }
      }
    }

    if (table_idx < share->table_count)
    {
      correspond_flag = TRUE;
      tmp_columns_bit =
        &share->correspond_columns_bit[table_idx * share->bitmap_size];
      for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
      {
        work_bitmap[roop_count] =
          read_set[roop_count] & tmp_columns_bit[roop_count];
        read_set[roop_count] &= ~work_bitmap[roop_count];

        work_bitmap2[roop_count] =
          write_set[roop_count] & tmp_columns_bit[roop_count];
        write_set[roop_count] &= ~work_bitmap2[roop_count];

        if (read_set[roop_count] || write_set[roop_count])
          correspond_flag = FALSE;
      }
      if (!set_used_table)
      {
        child_table_idx = table_idx;
        if (correspond_flag)
        {
          single_table = TRUE;
        }
      }
      set_used_table = TRUE;
      clear_child_bitmap(table_idx);
      set_child_bitmap(work_bitmap, table_idx, FALSE);
      set_child_bitmap(work_bitmap2, table_idx, TRUE);
      vp_set_bit(use_tables, table_idx);
      if (correspond_flag)
        break;
    }
  } while (table_idx < share->table_count);
  for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
  {
    if (vp_bit_count(read_set[roop_count] | write_set[roop_count]))
    {
      my_printf_error(ER_VP_IGNORED_CORRESPOND_COLUMN_NUM,
        ER_VP_IGNORED_CORRESPOND_COLUMN_STR, MYF(0));
      DBUG_RETURN(ER_VP_IGNORED_CORRESPOND_COLUMN_NUM);
    }
  }
  DBUG_RETURN(0);
}

void ha_vp::clear_child_bitmap(
  int table_idx
) {
  TABLE *child_table = part_tables[table_idx].table;
  int bitmap_size = (child_table->s->fields + 7) / 8;
  DBUG_ENTER("ha_vp::clear_child_bitmap");
  memset((uchar *) child_table->write_set->bitmap, 0,
    sizeof(uchar) * bitmap_size);
  memset((uchar *) child_table->read_set->bitmap, 0,
    sizeof(uchar) * bitmap_size);
  DBUG_VOID_RETURN;
}

uchar *ha_vp::create_child_key(
  const uchar *key_same,
  uchar *key_different,
  key_part_map keypart_map
) {
  int roop_count, length;
  uint store_length;
  uchar *ptr;
  TABLE *table2;
  Field *field, *field2;
  KEY *key_info, *key_info2;
  KEY_PART_INFO *key_part, *key_part2;
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_bin), str2;
  key_part_map tmp_key_part_map;
  DBUG_ENTER("ha_vp::create_child_key");
  if (vp_bit_is_set(share->need_converting, child_table_idx))
  {
    table2 = part_tables[child_table_idx].table;
    key_info = &table->key_info[active_index];
    key_part = key_info->key_part;
    key_info2 = &table2->key_info[child_key_idx];
    key_part2 = key_info2->key_part;
    tmp_key_part_map = make_prev_keypart_map(key_info->key_parts);
    tmp_key_part_map &= keypart_map;
#ifndef DBUG_OFF
    my_bitmap_map *tmp_map_r = dbug_tmp_use_all_columns(table,
      table->read_set);
    my_bitmap_map *tmp_map_w2 = dbug_tmp_use_all_columns(table2,
      table2->write_set);
#endif
    for (
      roop_count = 0, ptr = (uchar *) key_same,
      store_length = key_part->store_length;
      tmp_key_part_map > 0;
      ptr += store_length, roop_count++, tmp_key_part_map >>= 1,
      store_length = key_part->store_length
    ) {
      field = key_part[roop_count].field;
      field2 = key_part2[roop_count].field;
      if (key_part[roop_count].null_bit && *ptr++)
      {
        field2->set_null();
        field2->reset();
      } else {
        if (
          field->type() == MYSQL_TYPE_BLOB ||
          field->real_type() == MYSQL_TYPE_VARCHAR ||
          field->type() == MYSQL_TYPE_GEOMETRY
        ) {
          length = uint2korr(ptr);
          str2.set_quick((char *) ptr + HA_KEY_BLOB_LENGTH, length,
            &my_charset_bin);
          field2->set_notnull();
          field2->store(
            length ? str2.ptr() : NullS, length,
            field->charset());
        } else {
          field->val_str(&str, ptr);
          length = str.length();
          field2->set_notnull();
          field2->store(
            length ? str.ptr() : NullS, length,
            field->charset());
        }
      }
    }
#ifndef DBUG_OFF
    dbug_tmp_restore_column_map(table->read_set, tmp_map_r);
    dbug_tmp_restore_column_map(table2->write_set, tmp_map_w2);
#endif
    key_copy(
      key_different,
      table2->record[0],
      key_info2,
      key_info2->key_length);
    DBUG_RETURN(key_different);
  }
  DBUG_RETURN((uchar *) key_same);
}

int ha_vp::get_child_record_by_idx(
  int table_idx,
  my_ptrdiff_t ptr_diff
) {
  int roop_count, length;
  TABLE *table2;
  uchar *tmp_bitmap, *tmp_bitmap2, *pk_bitmap;
  Field **field_ptr, **field_ptr2, *field, *field2;
  int *correspond_columns_c, column_idx;
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_bin);
  DBUG_ENTER("ha_vp::get_child_record_by_idx");
  if (table_idx == child_table_idx)
    pk_bitmap = NULL;
  else
    pk_bitmap = share->keys[table_share->primary_key].columns_bit;
  table2 = part_tables[table_idx].table;
  tmp_bitmap = (uchar *) table2->read_set->bitmap;
  tmp_bitmap2 = (uchar *) table2->write_set->bitmap;
  field_ptr = table->field;
  field_ptr2 = table2->field;
  correspond_columns_c = share->correspond_columns_c_ptr[table_idx];
#ifndef DBUG_OFF
  my_bitmap_map *tmp_map_w = dbug_tmp_use_all_columns(table, table->write_set);
#endif
  for (roop_count = 0; roop_count < table2->s->fields; roop_count++)
  {
    if (
      vp_bit_is_set(tmp_bitmap, roop_count) ||
      vp_bit_is_set(tmp_bitmap2, roop_count)
    ) {
#ifndef DBUG_OFF
      my_bitmap_map *tmp_map_r2 = dbug_tmp_use_all_columns(table2,
        table2->read_set);
#endif
      column_idx = correspond_columns_c[roop_count];
      if (!pk_bitmap || !(vp_bit_is_set(pk_bitmap, column_idx)))
      {
        DBUG_PRINT("info",("vp set field %d-%d to %d",
          child_table_idx, roop_count, column_idx));
        field2 = field_ptr2[roop_count];
        field = field_ptr[column_idx];
        field->move_field_offset(ptr_diff);
        if (field2->is_null())
        {
          field->set_null();
          field->reset();
        } else {
          field->set_notnull();
          if (field->flags & BLOB_FLAG)
          {
            DBUG_PRINT("info", ("vp blob field"));
            if (
              (field2->flags & BLOB_FLAG) &&
              (
                field->charset() == &my_charset_bin ||
                field->charset()->cset == field2->charset()->cset
              )
            ) {
              uchar *tmp_char;
              ((Field_blob *)field2)->get_ptr(&tmp_char);
              ((Field_blob *)field)->set_ptr(
                ((Field_blob *)field2)->get_length(), tmp_char);
            } else {
              DBUG_PRINT("info", ("vp blob convert"));
              String *str2 = &blob_buff[field->field_index];
              str2->length(0);
              field2->val_str(&str);
              if (str2->append(str.ptr(), str.length(), field2->charset()))
              {
#ifndef DBUG_OFF
                dbug_tmp_restore_column_map(table->write_set, tmp_map_w);
                dbug_tmp_restore_column_map(table2->read_set, tmp_map_r2);
#endif
                DBUG_RETURN(HA_ERR_OUT_OF_MEM);
              }
              ((Field_blob *)field)->set_ptr(str2->length(),
                (uchar *) str2->ptr());
            }
          } else {
            field2->val_str(&str);
            length = str.length();
            field->store(
              length ? str.ptr() : NullS, length,
              field2->charset());
          }
        }
        field->move_field_offset(-ptr_diff);
      }
#ifndef DBUG_OFF
      dbug_tmp_restore_column_map(table2->read_set, tmp_map_r2);
#endif
    }
  }
#ifndef DBUG_OFF
  dbug_tmp_restore_column_map(table->write_set, tmp_map_w);
#endif
  DBUG_RETURN(0);
}

int ha_vp::get_child_record_by_pk(
  my_ptrdiff_t ptr_diff
) {
  int error_num = 0, roop_count;
  uchar *table_key;
  uchar table_key_different[MAX_KEY_LENGTH];
  TABLE *table2;
  VP_KEY_COPY vp_key_copy;
#ifndef WITHOUT_VP_BG_ACCESS
  int bgs_mode =
    THDVAR(table->in_use, bgs_mode) < 0 ?
    share->bgs_mode :
    THDVAR(table->in_use, bgs_mode);
  VP_BG_BASE *base;
#endif
  DBUG_ENTER("ha_vp::get_child_record_by_pk");
  vp_key_copy.init = FALSE;
  vp_key_copy.mem_root_init = FALSE;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      child_table_idx != roop_count &&
      vp_bit_is_set(use_tables, roop_count)
    ) {
      table2 = part_tables[roop_count].table;
#ifndef WITHOUT_VP_BG_ACCESS
      if (
        bgs_mode &&
        (table2->file->ha_table_flags() & VP_CAN_BG_SEARCH)
      ) {
        base = &bg_base[roop_count];
        vp_key_copy.table_key_different = base->table_key_different;
        if (
          (error_num = create_bg_thread(base)) ||
          (error_num = search_by_pk(roop_count, 0, &vp_key_copy, 0,
            (uchar **) &base->table_key))
        )
          goto error;
        base->tgt_key_part_map = vp_key_copy.tgt_key_part_map;
        base->bg_command = VP_BG_COMMAND_SELECT;
        bg_kick(base);
      } else {
#endif
        vp_key_copy.table_key_different = table_key_different;
        if (
          (error_num = search_by_pk(roop_count, 0, &vp_key_copy, 0,
            &table_key)) ||
          (error_num = table2->file->index_read_map(
            table2->record[0], table_key, vp_key_copy.tgt_key_part_map,
            HA_READ_KEY_EXACT))
        )
          goto error;
#ifndef WITHOUT_VP_BG_ACCESS
      }
#endif
    }
  }
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      child_table_idx != roop_count &&
      vp_bit_is_set(use_tables, roop_count)
    ) {
#ifndef WITHOUT_VP_BG_ACCESS
      table2 = part_tables[roop_count].table;
      if (
        bgs_mode &&
        (table2->file->ha_table_flags() & VP_CAN_BG_SEARCH)
      ) {
        base = &bg_base[roop_count];
        bg_wait(base);
        if (base->bg_error)
        {
          error_num = base->bg_error;
          goto error;
        }
      }
#endif
      if ((error_num = get_child_record_by_idx(roop_count, ptr_diff)))
        goto error;
    }
  }

  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(error_num);

error:
#ifndef WITHOUT_VP_BG_ACCESS
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    base = &bg_base[roop_count];
    if (base->bg_init)
      bg_wait(base);
  }
#endif
  if (vp_key_copy.mem_root_init)
    free_root(&vp_key_copy.mem_root, MYF(0));
  DBUG_RETURN(error_num);
}

bool ha_vp::set_child_bitmap(
  uchar *bitmap,
  int table_idx,
  bool write_flg
) {
  bool ret_flag = FALSE;
  int roop_count, field_idx;
  uchar *tmp_bitmap, *pk_bitmap;
  int *correspond_columns_p =
    &share->correspond_columns_p[table_idx * table_share->fields];
  TABLE *child_table = part_tables[table_idx].table;
  bool use_full_column = FALSE;
  DBUG_ENTER("ha_vp::set_child_bitmap");
  DBUG_PRINT("info",("vp table_idx=%d", table_idx));
  DBUG_PRINT("info",("vp write_flg=%d", write_flg));
  if (write_flg)
    tmp_bitmap = (uchar *) child_table->write_set->bitmap;
  else
    tmp_bitmap = (uchar *) child_table->read_set->bitmap;

  if (
    !write_flg &&
    update_request &&
    share->need_full_col_for_update[table_idx]
  ) {
    use_full_column = TRUE;
    child_keyread = FALSE;
  }

  if (
    !use_full_column &&
    !write_flg &&
    table_idx == child_table_idx &&
    (!single_table || update_request)
  ) {
    pk_bitmap = share->keys[table_share->primary_key].columns_bit;
    if (child_keyread)
    {
      VP_CORRESPOND_KEY *correspond_key =
        share->correspond_keys_p_ptr[child_table_idx];
      uchar *key_bitmap = correspond_key[child_key_idx].columns_bit;
      for (roop_count = 0; roop_count < share->bitmap_size; roop_count++)
      {
        if ((key_bitmap[roop_count] & pk_bitmap[roop_count]) !=
          pk_bitmap[roop_count])
        {
          DBUG_PRINT("info",("vp cancel keyread"));
          child_keyread = FALSE;
          break;
        }
      }
    }
  } else
    pk_bitmap = NULL;

  for (roop_count = 0; roop_count < table_share->fields; roop_count++)
  {
    if (
      use_full_column ||
      vp_bit_is_set(bitmap, roop_count) ||
      (pk_bitmap && vp_bit_is_set(pk_bitmap, roop_count))
    ) {
      if ((field_idx = correspond_columns_p[roop_count]) < MAX_FIELDS)
      {
        vp_set_bit(tmp_bitmap, field_idx);
        ret_flag = TRUE;
      }
    }
  }
#ifndef DBUG_OFF
  int bitmap_size = (child_table->s->fields + 7) / 8;
  for (roop_count = 0; roop_count < bitmap_size; roop_count++)
    DBUG_PRINT("info",("vp bitmap is %d-%d", roop_count,
      tmp_bitmap[roop_count]));
#endif
  DBUG_RETURN(ret_flag);
}

void ha_vp::set_child_record_for_update(
  my_ptrdiff_t ptr_diff,
  int record_idx,
  bool write_flg,
  bool use_table_chk
) {
  int roop_count, roop_count2, length,
    *correspond_columns_c, column_idx;
  my_ptrdiff_t ptr_diff2;
  TABLE *table2;
  Field **field_ptr, **field_ptr2, *field, *field2;
  MY_BITMAP *my_bitmap;
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_bin);
  DBUG_ENTER("ha_vp::set_child_record_for_update");
  field_ptr = table->field;
  for (roop_count = 0; roop_count < share->table_count; roop_count++)
  {
    if (
      use_table_chk ||
      vp_bit_is_set(use_tables2, roop_count)
    ) {
      table2 = part_tables[roop_count].table;
      field_ptr2 = table2->field;
      correspond_columns_c = share->correspond_columns_c_ptr[roop_count];
      if (write_flg)
        my_bitmap = table2->write_set;
      else
        my_bitmap = table2->read_set;
#ifndef DBUG_OFF
      my_bitmap_map *tmp_map_r = dbug_tmp_use_all_columns(table,
        table->read_set);
      my_bitmap_map *tmp_map_w2;
      if (!write_flg)
        tmp_map_w2 = dbug_tmp_use_all_columns(table2,
          table2->write_set);
#endif
      ptr_diff2 = PTR_BYTE_DIFF(table2->record[record_idx], table2->record[0]);
      for (roop_count2 = 0; roop_count2 < table2->s->fields; roop_count2++)
      {
        column_idx = correspond_columns_c[roop_count2];
        field2 = field_ptr2[roop_count2];
        if (bitmap_is_set(my_bitmap, roop_count2))
        {
          DBUG_PRINT("info",("vp set field %d to %d-%d",
            column_idx, roop_count, roop_count2));
          field = field_ptr[column_idx];
          field->move_field_offset(ptr_diff);
          field2->move_field_offset(ptr_diff2);
          if (field->is_null())
          {
            field2->set_null();
            field2->reset();
          } else {
            field->val_str(&str);
            length = str.length();
            field2->set_notnull();
            field2->store(
              length ? str.ptr() : NullS, length,
              field->charset());
          }
          field2->move_field_offset(-ptr_diff2);
          field->move_field_offset(-ptr_diff);
          if (use_table_chk)
            vp_set_bit(use_tables2, roop_count);
        }
      }
#ifndef DBUG_OFF
      dbug_tmp_restore_column_map(table->read_set, tmp_map_r);
      if (!write_flg)
        dbug_tmp_restore_column_map(table2->write_set, tmp_map_w2);
#endif
    }
  }

  DBUG_VOID_RETURN;
}

int ha_vp::search_by_pk(
  int table_idx,
  int record_idx,
  VP_KEY_COPY *vp_key_copy,
  my_ptrdiff_t ptr_diff,
  uchar **table_key
) {
  int error_num, roop_count, key_idx;
  TABLE *table2;
  KEY *key_info, *key_info2;
  Field *field, *field2;
  KEY_PART_INFO *key_part, *key_part2;
  char buff[MAX_FIELD_WIDTH];
  String str(buff, sizeof(buff), &my_charset_bin);
  DBUG_ENTER("ha_vp::search_by_pk");
  table2 = part_tables[table_idx].table;
  key_idx = share->correspond_pk[table_idx]->key_idx;
/*
  my_ptrdiff_t ptr_diff2 =
    PTR_BYTE_DIFF(table2->record[record_idx], table2->record[0]);
*/
  key_info = &table->key_info[table_share->primary_key];
  key_part = key_info->key_part;

  if (!vp_key_copy->init)
  {
    vp_key_copy->init = TRUE;
    for (roop_count = 0; roop_count < key_info->key_parts; roop_count++)
    {
      field = key_part[roop_count].field;
      field->move_field_offset(ptr_diff);
    }
    key_copy(
      vp_key_copy->table_key_same,
      table->record[record_idx],
      key_info,
      key_info->key_length);
    for (roop_count = 0; roop_count < key_info->key_parts; roop_count++)
    {
      field = key_part[roop_count].field;
      field->move_field_offset(-ptr_diff);
    }
    vp_key_copy->tgt_key_part_map = make_prev_keypart_map(key_info->key_parts);
  }

  if (vp_bit_is_set(share->need_converting, table_idx))
  {
#ifndef DBUG_OFF
    my_bitmap_map *tmp_map_r = dbug_tmp_use_all_columns(table,
      table->read_set);
    my_bitmap_map *tmp_map_w2 = dbug_tmp_use_all_columns(table2,
      table2->write_set);
#endif
    if (!vp_key_copy->mem_root_init)
    {
      vp_key_copy->mem_root_init = TRUE;
      init_alloc_root(&vp_key_copy->mem_root, 1024, 0);
      if (
        !(vp_key_copy->ptr = (char **) sql_alloc(
          sizeof(char *) * key_info->key_parts)) ||
        !(vp_key_copy->len = (int *) sql_alloc(
          sizeof(int) * key_info->key_parts)) ||
        !(vp_key_copy->null_flg = (uchar *) sql_alloc(
          sizeof(uchar) * ((key_info->key_parts + 7) / 8)))
      ) {
#ifndef DBUG_OFF
        dbug_tmp_restore_column_map(table->read_set, tmp_map_r);
        dbug_tmp_restore_column_map(table2->write_set, tmp_map_w2);
#endif
        DBUG_RETURN(HA_ERR_OUT_OF_MEM);
      }
      memset((uchar *) vp_key_copy->null_flg, 0,
        sizeof(uchar) * ((key_info->key_parts + 7) / 8));
      for (roop_count = 0; roop_count < key_info->key_parts;
        roop_count++)
      {
        field = key_part[roop_count].field;
        field->move_field_offset(ptr_diff);
        if (field->is_null())
        {
          vp_set_bit(vp_key_copy->null_flg, roop_count);
        } else {
          field->val_str(&str);
          vp_key_copy->len[roop_count] = str.length();
          if (!vp_key_copy->len[roop_count])
            vp_key_copy->ptr[roop_count] = NullS;
          else if ((vp_key_copy->ptr[roop_count] =
            (char *) alloc_root(&vp_key_copy->mem_root,
              vp_key_copy->len[roop_count])))
          {
            memcpy(vp_key_copy->ptr[roop_count], str.ptr(),
              vp_key_copy->len[roop_count]);
          } else {
#ifndef DBUG_OFF
            dbug_tmp_restore_column_map(table->read_set, tmp_map_r);
            dbug_tmp_restore_column_map(table2->write_set, tmp_map_w2);
#endif
            DBUG_RETURN(HA_ERR_OUT_OF_MEM);
          }
        }
        field->move_field_offset(-ptr_diff);
      }
    }

    key_info2 = &table2->key_info[key_idx];
    key_part2 = key_info2->key_part;
    for (roop_count = 0; roop_count < key_info->key_parts; roop_count++)
    {
      field2 = key_part2[roop_count].field;
/*
      field2->move_field_offset(ptr_diff2);
*/
      if (vp_bit_is_set(vp_key_copy->null_flg, roop_count))
      {
        field2->set_null();
        field2->reset();
      } else {
        field2->set_notnull();
        field2->store(
          vp_key_copy->ptr[roop_count], vp_key_copy->len[roop_count],
          key_part[roop_count].field->charset());
      }
    }
#ifndef DBUG_OFF
    dbug_tmp_restore_column_map(table->read_set, tmp_map_r);
    dbug_tmp_restore_column_map(table2->write_set, tmp_map_w2);
#endif

    *table_key = vp_key_copy->table_key_different;
    key_copy(
      vp_key_copy->table_key_different,
      table2->record[0],
      key_info2,
      key_info2->key_length);

/*
    for (roop_count = 0; roop_count < key_info->key_parts; roop_count++)
    {
      field2 = key_part2[roop_count].field;
      field2->move_field_offset(-ptr_diff2);
    }
*/
  } else
    *table_key = vp_key_copy->table_key_same;
  DBUG_RETURN(0);
}

int ha_vp::search_by_pk_for_update(
  int table_idx,
  int record_idx,
  VP_KEY_COPY *vp_key_copy,
  my_ptrdiff_t ptr_diff,
  int bgu_mode
) {
  int error_num, key_idx;
  TABLE *table2;
  uchar *table_key;
  uchar table_key_different[MAX_KEY_LENGTH];
#ifndef WITHOUT_VP_BG_ACCESS
  VP_BG_BASE *base;
#endif
  DBUG_ENTER("ha_vp::search_by_pk_for_update");
  table2 = part_tables[table_idx].table;
  key_idx = share->correspond_pk[table_idx]->key_idx;

#ifndef WITHOUT_VP_BG_ACCESS
  if (
    bgu_mode &&
    (table2->file->ha_table_flags() & VP_CAN_BG_UPDATE)
  ) {
    base = &bg_base[table_idx];
    vp_key_copy->table_key_different = base->table_key_different;
    if (
      (error_num = create_bg_thread(base)) ||
      (error_num = search_by_pk(table_idx, record_idx, vp_key_copy, ptr_diff,
        (uchar **) &base->table_key))
    )
      DBUG_RETURN(error_num);
    base->tgt_key_part_map = vp_key_copy->tgt_key_part_map;
    base->key_idx = key_idx;
    base->record_idx = record_idx;
    base->bg_command = VP_BG_COMMAND_UPDATE_SELECT;
    bg_kick(base);
  } else {
#endif
    vp_key_copy->table_key_different = table_key_different;
    if ((error_num = search_by_pk(table_idx, record_idx, vp_key_copy, ptr_diff,
      &table_key)))
      DBUG_RETURN(error_num);

    if (
      !(error_num = table2->file->index_read_idx_map(
        table2->record[0], key_idx, table_key,
        vp_key_copy->tgt_key_part_map, HA_READ_KEY_EXACT)) &&
      record_idx
    ) {
      store_record(table2, record[1]);
    }
#ifndef WITHOUT_VP_BG_ACCESS
  }
#endif
  DBUG_RETURN(error_num);
}

int ha_vp::create_bg_thread(
  VP_BG_BASE *base
) {
  int error_num;
  DBUG_ENTER("ha_vp::create_bg_thread");
  if (!base->bg_init)
  {
    if (pthread_mutex_init(&base->bg_sync_mutex, MY_MUTEX_INIT_FAST))
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_sync_mutex_init;
    }
    if (pthread_mutex_init(&base->bg_mutex, MY_MUTEX_INIT_FAST))
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_mutex_init;
    }
    if (pthread_cond_init(&base->bg_sync_cond, NULL))
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_sync_cond_init;
    }
    if (pthread_cond_init(&base->bg_cond, NULL))
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_cond_init;
    }
    pthread_mutex_lock(&base->bg_mutex);
    if (pthread_create(&base->bg_thread, &vp_pt_attr,
      vp_bg_action, (void *) base)
    ) {
      pthread_mutex_unlock(&base->bg_mutex);
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_thread_create;
    }
    pthread_cond_wait(&base->bg_cond, &base->bg_mutex);
    pthread_mutex_unlock(&base->bg_mutex);
    if (!base->bg_init)
    {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_thread_create;
    }
  }
  DBUG_RETURN(0);

error_thread_create:
  VOID(pthread_cond_destroy(&base->bg_cond));
error_cond_init:
  VOID(pthread_cond_destroy(&base->bg_sync_cond));
error_sync_cond_init:
  VOID(pthread_mutex_destroy(&base->bg_mutex));
error_mutex_init:
  VOID(pthread_mutex_destroy(&base->bg_sync_mutex));
error_sync_mutex_init:
  DBUG_RETURN(error_num);
}

void ha_vp::free_bg_thread(
  VP_BG_BASE *base
) {
  DBUG_ENTER("ha_vp::free_bg_thread");
  if (base->bg_init)
  {
    pthread_mutex_lock(&base->bg_mutex);
    base->bg_command = VP_BG_COMMAND_KILL;
    pthread_cond_signal(&base->bg_cond);
    pthread_cond_wait(&base->bg_cond, &base->bg_mutex);
    pthread_mutex_unlock(&base->bg_mutex);
    VOID(pthread_cond_destroy(&base->bg_cond));
    VOID(pthread_cond_destroy(&base->bg_sync_cond));
    VOID(pthread_mutex_destroy(&base->bg_mutex));
    VOID(pthread_mutex_destroy(&base->bg_sync_mutex));
    base->bg_init = FALSE;
  }
  DBUG_VOID_RETURN;
}

void ha_vp::bg_kick(
  VP_BG_BASE *base
) {
  DBUG_ENTER("ha_vp::bg_kick");
  pthread_mutex_lock(&base->bg_mutex);
  base->bg_caller_sync_wait = TRUE;
  pthread_mutex_lock(&base->bg_sync_mutex);
  pthread_cond_signal(&base->bg_cond);
  pthread_mutex_unlock(&base->bg_mutex);
  pthread_cond_wait(&base->bg_sync_cond, &base->bg_sync_mutex);
  pthread_mutex_unlock(&base->bg_sync_mutex);
  base->bg_caller_sync_wait = FALSE;
  DBUG_VOID_RETURN;
}

void ha_vp::bg_wait(
  VP_BG_BASE *base
) {
  DBUG_ENTER("ha_vp::bg_wait");
  pthread_mutex_lock(&base->bg_mutex);
  pthread_mutex_unlock(&base->bg_mutex);
  DBUG_VOID_RETURN;
}
